# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-

import functools

import brainstate
import braintools
import jax
import jax.numpy as jnp
import pytest

from brainevent._fcn.float import fcnmv, fcnmv_p, fcnmm_p, fcnmm_p_call
from brainevent._test_util import (
    generate_fixed_conn_num_indices,
    vector_fcn,
    matrix_fcn,
    fcn_vector,
    fcn_matrix,
    allclose,
    ones_like,
)

platform = jax.default_backend()
FCNMV_IMPLEMENTATIONS = tuple(fcnmv_p.available_backends(platform))
FCNMM_IMPLEMENTATIONS = tuple(fcnmm_p.available_backends(platform))

if platform == 'cpu':
    shapes = [
        (20, 40),
        (50, 30),
    ]
else:
    shapes = [
        (20, 40),
        # (50, 30),
        # (200, 400),
        (500, 300),
        # (2000, 4000),
        (5000, 3000),
    ]


def _make_data(homo_w, shape):
    if homo_w:
        return jnp.asarray(1.5, dtype=jnp.float32)
    return braintools.init.Normal(0.0, 1.0)(shape)


def _vector_fcn_api(x, data, indices, shape, implementation):
    return fcnmv(
        data,
        indices,
        x,
        shape=shape,
        transpose=True,
        backend=implementation,
    )


def _fcn_vector_api(x, data, indices, shape, implementation):
    return fcnmv(
        data,
        indices,
        x,
        shape=shape,
        transpose=False,
        backend=implementation,
    )


def _matrix_fcn_api(x, data, indices, shape, implementation):
    return fcnmm_p_call(
        data,
        indices,
        x.T,
        shape=shape,
        transpose=True,
        backend=implementation,
    )[0].T


def _fcn_matrix_api(x, data, indices, shape, implementation):
    return fcnmm_p_call(
        data,
        indices,
        x,
        shape=shape,
        transpose=False,
        backend=implementation,
    )[0]


@pytest.mark.skipif(
    not FCNMV_IMPLEMENTATIONS,
    reason=f'No fcnmv implementation on platform={platform}',
)
@pytest.mark.parametrize('implementation', FCNMV_IMPLEMENTATIONS)
class TestVector:
    def _generate_x(self, shape, require_float=False):
        if isinstance(shape, (tuple, list)):
            yield brainstate.random.rand(*shape)
        else:
            yield brainstate.random.rand(shape)

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    def test_vector_csr(self, implementation, homo_w, shape):
        m, n = shape
        for x in self._generate_x(m):
            indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))
            data = _make_data(homo_w, indices.shape)
            y = _vector_fcn_api(x, data, indices, (m, n), implementation)
            y_true = vector_fcn(x, data, indices, (m, n))
            assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
            jax.block_until_ready((x, indices, data, y, y_true))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    def test_csr_vector(self, implementation, homo_w, shape):
        m, n = shape
        for v in self._generate_x(n):
            indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))
            data = _make_data(homo_w, indices.shape)
            y = _fcn_vector_api(v, data, indices, (m, n), implementation)
            y_true = fcn_vector(v, data, indices, (m, n))
            assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
            jax.block_until_ready((v, indices, data, y, y_true))

    @pytest.mark.parametrize('transpose', [True, False])
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    def test_vjp(self, implementation, transpose, homo_w, shape):
        n_in, n_out = shape
        indices = generate_fixed_conn_num_indices(n_in, n_out, int(n_out * 0.1))
        w = _make_data(homo_w, indices.shape)

        def f_api(x, data):
            if transpose:
                r = _vector_fcn_api(x, data, indices, shape, implementation)
            else:
                r = _fcn_vector_api(x, data, indices, shape, implementation)
            return r.sum()

        def f_ref(x, data):
            if transpose:
                r = vector_fcn(x, data, indices, shape)
            else:
                r = fcn_vector(x, data, indices, shape)
            return r.sum()

        for x in self._generate_x(n_in if transpose else n_out, require_float=True):
            r1 = jax.jit(lambda x, data: jax.grad(f_api, argnums=(0, 1))(x, data))(x, w)
            r2 = jax.jit(lambda x, data: jax.grad(f_ref, argnums=(0, 1))(x, data))(x, w)
            assert allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)
            jax.block_until_ready((x, indices, w, r1[0], r1[1], r2[0], r2[1]))

    @pytest.mark.parametrize('transpose', [True, False])
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    def test_jvp(self, implementation, transpose, homo_w, shape):
        n_in, n_out = shape
        indices = generate_fixed_conn_num_indices(n_in, n_out, int(n_out * 0.1))
        w = _make_data(homo_w, indices.shape)

        def f_api(x, data):
            if transpose:
                return _vector_fcn_api(x, data, indices, shape, implementation)
            return _fcn_vector_api(x, data, indices, shape, implementation)

        def f_ref(x, data):
            if transpose:
                return vector_fcn(x, data, indices, shape)
            return fcn_vector(x, data, indices, shape)

        for x in self._generate_x(n_in if transpose else n_out, require_float=True):
            o1, r1 = jax.jit(
                lambda x, data: jax.jvp(f_api, (x, data), (ones_like(x), ones_like(data)))
            )(x, w)
            o2, r2 = jax.jit(
                lambda x, data: jax.jvp(f_ref, (x, data), (ones_like(x), ones_like(data)))
            )(x, w)
            assert allclose(r1, r2, rtol=1e-3, atol=1e-3)
            assert allclose(o1, o2, rtol=1e-3, atol=1e-3)
            jax.block_until_ready((x, indices, w, o1, r1, o2, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('batch_size', [32])
    def test_batching_weight(self, implementation, homo_w, shape, batch_size):
        m, n = shape
        indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))
        data = (
            brainstate.random.rand(batch_size)
            if homo_w else
            braintools.init.Normal(0., 1.)((batch_size,) + indices.shape)
        )

        x_left = brainstate.random.rand(m)
        y = jax.jit(jax.vmap(
            lambda w: _vector_fcn_api(x_left, w, indices, (m, n), implementation)
        ))(data)
        y_true = jax.jit(jax.vmap(
            lambda w: vector_fcn(x_left, w, indices, (m, n))
        ))(data)
        assert allclose(y, y_true, rtol=1e-3, atol=1e-3)

        x_right = brainstate.random.rand(n)
        y = jax.jit(jax.vmap(
            lambda w: _fcn_vector_api(x_right, w, indices, (m, n), implementation)
        ))(data)
        y_true = jax.jit(jax.vmap(
            lambda w: fcn_vector(x_right, w, indices, (m, n))
        ))(data)
        assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((indices, data, x_left, x_right, y, y_true))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('batch_size', [32])
    @pytest.mark.parametrize('batch_axis', [0, 1])
    def test_batching_vector(self, implementation, homo_w, shape, batch_size, batch_axis):
        m, n = shape
        indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))
        data = _make_data(homo_w, indices.shape)

        @jax.jit
        @functools.partial(jax.vmap, in_axes=batch_axis)
        def f_compare_vector(x):
            y = _vector_fcn_api(x, data, indices, (m, n), implementation)
            y_true = vector_fcn(x, data, indices, (m, n))
            return y, y_true

        xs = (
            brainstate.random.rand(batch_size, m)
            if batch_axis == 0 else
            brainstate.random.rand(m, batch_size)
        )
        y, y_true = f_compare_vector(xs)
        assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((indices, data, xs, y, y_true))


@pytest.mark.skipif(
    not FCNMM_IMPLEMENTATIONS,
    reason=f'No fcnmm implementation on platform={platform}',
)
@pytest.mark.parametrize('implementation', FCNMM_IMPLEMENTATIONS)
class TestMatrix:
    def _generate_x(self, shape, require_float=False):
        if isinstance(shape, (tuple, list)):
            yield brainstate.random.rand(*shape)
        else:
            yield brainstate.random.rand(shape)

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('k', [10])
    def test_matrix_csr(self, implementation, homo_w, shape, k):
        m, n = shape
        for x in self._generate_x([k, m]):
            indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))
            data = _make_data(homo_w, indices.shape)
            y = _matrix_fcn_api(x, data, indices, (m, n), implementation)
            y_true = matrix_fcn(x, data, indices, (m, n))
            assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
            jax.block_until_ready((x, indices, data, y, y_true))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('k', [10])
    def test_csr_matrix(self, implementation, homo_w, shape, k):
        m, n = shape
        for matrix in self._generate_x([n, k]):
            indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))
            data = _make_data(homo_w, indices.shape)
            y = _fcn_matrix_api(matrix, data, indices, (m, n), implementation)
            y_true = fcn_matrix(matrix, data, indices, (m, n))
            assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
            jax.block_until_ready((matrix, indices, data, y, y_true))

    @pytest.mark.parametrize('transpose', [True, False])
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('k', [10])
    def test_vjp(self, implementation, transpose, homo_w, shape, k):
        n_in, n_out = shape
        indices = generate_fixed_conn_num_indices(n_in, n_out, int(n_out * 0.1))
        w = _make_data(homo_w, indices.shape)

        def f_api(x, data):
            if transpose:
                r = _matrix_fcn_api(x, data, indices, shape, implementation)
            else:
                r = _fcn_matrix_api(x, data, indices, shape, implementation)
            return r.sum()

        def f_ref(x, data):
            if transpose:
                r = matrix_fcn(x, data, indices, shape)
            else:
                r = fcn_matrix(x, data, indices, shape)
            return r.sum()

        for x in self._generate_x([k, n_in] if transpose else [n_out, k], require_float=True):
            r1 = jax.jit(lambda x, data: jax.grad(f_api, argnums=(0, 1))(x, data))(x, w)
            r2 = jax.jit(lambda x, data: jax.grad(f_ref, argnums=(0, 1))(x, data))(x, w)
            assert allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)
            jax.block_until_ready((x, indices, w, r1[0], r1[1], r2[0], r2[1]))

    @pytest.mark.parametrize('transpose', [True, False])
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('k', [10])
    def test_jvp(self, implementation, transpose, homo_w, shape, k):
        n_in, n_out = shape
        indices = generate_fixed_conn_num_indices(n_in, n_out, int(n_out * 0.1))
        w = _make_data(homo_w, indices.shape)

        def f_api(x, data):
            if transpose:
                return _matrix_fcn_api(x, data, indices, shape, implementation)
            return _fcn_matrix_api(x, data, indices, shape, implementation)

        def f_ref(x, data):
            if transpose:
                return matrix_fcn(x, data, indices, shape)
            return fcn_matrix(x, data, indices, shape)

        for x in self._generate_x((k, n_in) if transpose else (n_out, k), require_float=True):
            o1, r1 = jax.jit(
                lambda x, data: jax.jvp(f_api, (x, data), (ones_like(x), ones_like(data)))
            )(x, w)
            o2, r2 = jax.jit(
                lambda x, data: jax.jvp(f_ref, (x, data), (ones_like(x), ones_like(data)))
            )(x, w)
            assert allclose(r1, r2, rtol=1e-3, atol=1e-3)
            assert allclose(o1, o2, rtol=1e-3, atol=1e-3)
            jax.block_until_ready((x, indices, w, o1, r1, o2, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('batch_size', [32])
    @pytest.mark.parametrize('k', [32])
    def test_batching_weight(self, implementation, homo_w, shape, batch_size, k):
        m, n = shape
        indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))

        data = (
            brainstate.random.rand(batch_size)
            if homo_w else
            braintools.init.Normal(0., 1.)((batch_size,) + indices.shape)
        )

        x_left = brainstate.random.rand(k, m)
        y = jax.jit(jax.vmap(
            lambda w: _matrix_fcn_api(x_left, w, indices, (m, n), implementation)
        ))(data)
        y_true = jax.jit(jax.vmap(
            lambda w: matrix_fcn(x_left, w, indices, (m, n))
        ))(data)
        assert allclose(y, y_true, rtol=1e-3, atol=1e-3)

        x_right = brainstate.random.rand(n, k)
        y = jax.jit(jax.vmap(
            lambda w: _fcn_matrix_api(x_right, w, indices, (m, n), implementation)
        ))(data)
        y_true = jax.jit(jax.vmap(
            lambda w: fcn_matrix(x_right, w, indices, (m, n))
        ))(data)
        assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((indices, data, x_left, x_right, y, y_true))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('batch_size', [32])
    @pytest.mark.parametrize('k', [32])
    @pytest.mark.parametrize('batch_axis', [0, 1, 2])
    def test_batching_vector(self, implementation, homo_w, shape, batch_size, k, batch_axis):
        m, n = shape
        indices = generate_fixed_conn_num_indices(m, n, int(n * 0.1))
        data = _make_data(homo_w, indices.shape)

        @jax.jit
        @functools.partial(jax.vmap, in_axes=batch_axis)
        def f_compare_matrix(x):
            y = _matrix_fcn_api(x, data, indices, (m, n), implementation)
            y_true = matrix_fcn(x, data, indices, (m, n))
            return y, y_true

        if batch_axis == 0:
            x_shape = [batch_size, k, m]
        elif batch_axis == 1:
            x_shape = [k, batch_size, m]
        else:
            x_shape = [k, m, batch_size]
        xs = brainstate.random.rand(*x_shape)
        y, y_true = f_compare_matrix(xs)
        assert allclose(y, y_true, rtol=1e-3, atol=1e-3)

        @jax.jit
        @functools.partial(jax.vmap, in_axes=batch_axis)
        def f_compare_fcn_matrix(x):
            y = _fcn_matrix_api(x, data, indices, (m, n), implementation)
            y_true = fcn_matrix(x, data, indices, (m, n))
            return y, y_true

        if batch_axis == 0:
            x_shape = [batch_size, n, k]
        elif batch_axis == 1:
            x_shape = [n, batch_size, k]
        else:
            x_shape = [n, k, batch_size]
        xs = brainstate.random.rand(*x_shape)
        y, y_true = f_compare_fcn_matrix(xs)
        assert allclose(y, y_true, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((indices, data, xs, y, y_true))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-


import brainstate
import braintools
import jax
import jax.numpy as jnp
import numpy as np
import pytest

from brainevent._fcn.binary import binary_fcnmv, binary_fcnmv_p, binary_fcnmm, binary_fcnmm_p
from brainevent._test_util import generate_fixed_conn_num_indices

platform = jax.default_backend()
FCNMV_IMPLEMENTATIONS = tuple(binary_fcnmv_p.available_backends(platform))
FCNMM_IMPLEMENTATIONS = tuple(binary_fcnmm_p.available_backends(platform))

if platform == 'cpu':
    SHAPES = (
        (20, 40),
        (50, 30)
    )
else:
    SHAPES = (
        (20, 40),
        # (50, 30),
        # (200, 400)
        (400, 200)
    )


def _implementation_params(implementations, op_name: str):
    if implementations:
        return [pytest.param(impl, id=impl) for impl in implementations]
    return [
        pytest.param(
            None,
            marks=pytest.mark.skip(reason=f'No {op_name} implementations on platform={platform}'),
            id=f'no-{op_name}',
        )
    ]


FCNMV_PARAMS = _implementation_params(FCNMV_IMPLEMENTATIONS, 'binary_fcnmv')
FCNMM_PARAMS = _implementation_params(FCNMM_IMPLEMENTATIONS, 'binary_fcnmm')


def _make_weights(indices, homo_w: bool):
    if homo_w:
        return jnp.asarray([1.5], dtype=jnp.float32)
    return braintools.init.Normal(0.0, 1.0)(indices.shape)


def _to_binary_events(events, dtype):
    events = jnp.asarray(events)
    if events.dtype == jnp.bool_:
        return jnp.asarray(events, dtype=dtype)
    return jnp.asarray(events > 0, dtype=dtype)


def _dense_from_fixed_conn(weights, indices, shape):
    n_pre, _ = shape
    rows = jnp.repeat(jnp.arange(n_pre, dtype=indices.dtype), indices.shape[1])
    cols = indices.reshape(-1)
    weights = jnp.asarray(weights)
    if weights.size == 1:
        value = jnp.reshape(weights, (-1,))[0]
        values = jnp.full((indices.size,), value, dtype=weights.dtype)
    else:
        values = weights.reshape(-1)
    return jnp.zeros(shape, dtype=weights.dtype).at[rows, cols].add(values)


def _mv_reference(weights, indices, events, shape, transpose):
    dense = _dense_from_fixed_conn(weights, indices, shape)
    events = _to_binary_events(events, dense.dtype)
    if transpose:
        return events @ dense
    return dense @ events


def _mm_reference(weights, indices, matrix, shape, transpose):
    dense = _dense_from_fixed_conn(weights, indices, shape)
    matrix = _to_binary_events(matrix, dense.dtype)
    if transpose:
        return jnp.matmul(dense.T, matrix, precision=jax.lax.Precision.HIGHEST)
    return jnp.matmul(dense, matrix, precision=jax.lax.Precision.HIGHEST)


@pytest.mark.parametrize('implementation', FCNMV_PARAMS)
@pytest.mark.parametrize('homo_w', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('event_dtype', [bool, float])
@pytest.mark.parametrize('shape', SHAPES)
def test_binary_fcnmv_forward_matches_reference(implementation, homo_w, transpose, event_dtype, shape):
    m, n = shape
    indices = generate_fixed_conn_num_indices(m, n, max(1, int(n * 0.1)))
    weights = _make_weights(indices, homo_w)

    event_size = m if transpose else n
    if event_dtype is bool:
        events = brainstate.random.rand(event_size) < 0.5
    else:
        raw = jnp.asarray(brainstate.random.rand(event_size), dtype=jnp.float32)
        events = jnp.where(raw > 0.4, raw, 0.0)

    y = binary_fcnmv(
        weights,
        indices,
        events,
        shape=shape,
        transpose=transpose,
        backend=implementation,
    )
    y_ref = _mv_reference(weights, indices, events, shape, transpose)
    assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)
    jax.block_until_ready((indices, weights, events, y, y_ref))


@pytest.mark.parametrize('implementation', FCNMM_PARAMS)
@pytest.mark.parametrize('homo_w', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('event_dtype', [bool, float])
@pytest.mark.parametrize('k', [10])
@pytest.mark.parametrize('shape', SHAPES)
def test_binary_fcnmm_forward_matches_reference(implementation, homo_w, transpose, event_dtype, k, shape):
    m, n = shape
    indices = generate_fixed_conn_num_indices(m, n, max(1, int(n * 0.1)))
    weights = _make_weights(indices, homo_w)

    n_rows = m if transpose else n
    if event_dtype is bool:
        matrix = brainstate.random.rand(n_rows, k) < 0.5
    else:
        raw = jnp.asarray(brainstate.random.rand(n_rows, k), dtype=jnp.float32)
        matrix = jnp.where(raw > 0.4, raw, 0.0)

    y = binary_fcnmm(
        weights,
        indices,
        matrix,
        shape=shape,
        transpose=transpose,
        backend=implementation,
    )
    y_ref = _mm_reference(weights, indices, matrix, shape, transpose)
    assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)
    jax.block_until_ready((indices, weights, matrix, y, y_ref))


@pytest.mark.parametrize('implementation', FCNMM_PARAMS)
@pytest.mark.parametrize('homo_w', [True, False])
@pytest.mark.parametrize('shape', [(20, 40), (50, 30)])
@pytest.mark.parametrize('k', [10])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_fcnmm_thresholds_float_events(implementation, homo_w, shape, k, transpose):
    m, n = shape
    indices = generate_fixed_conn_num_indices(m, n, max(1, int(n * 0.1)), replace=False)
    weights = _make_weights(indices, homo_w)

    rng = np.random.RandomState(42)
    n_rows = m if transpose else n
    raw = rng.rand(n_rows, k).astype(np.float32)
    mask = rng.rand(n_rows, k) > 0.5
    float_events = jnp.asarray(raw * mask)
    binary_events = jnp.asarray(float_events > 0, dtype=jnp.float32)

    y_float = binary_fcnmm(
        weights,
        indices,
        float_events,
        shape=shape,
        transpose=transpose,
        backend=implementation,
    )
    y_binary = binary_fcnmm(
        weights,
        indices,
        binary_events,
        shape=shape,
        transpose=transpose,
        backend=implementation,
    )
    assert jnp.allclose(y_float, y_binary, rtol=1e-3, atol=1e-3)
    jax.block_until_ready((indices, weights, float_events, binary_events, y_float, y_binary))



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# -*- coding: utf-8 -*-


import brainstate
import braintools
import jax
import jax.numpy as jnp
import pytest

from brainevent._csr.binary import binary_csrmv, binary_csrmv_p, binary_csrmm, binary_csrmm_p
from brainevent._csr.test_util import get_csr, vector_csr, matrix_csr, csr_vector, csr_matrix

platform = jax.default_backend()
CSRMV_IMPLEMENTATIONS = tuple(binary_csrmv_p.available_backends(platform))
CSRMM_IMPLEMENTATIONS = tuple(binary_csrmm_p.available_backends(platform))


def _require_implementations(implementations, op_name: str):
    if not implementations:
        pytest.skip(f'No {op_name} implementation on platform={platform}')


def _vector_csr_api(x, data, indices, indptr, shape, implementation):
    return binary_csrmv(
        data,
        indices,
        indptr,
        x,
        shape=shape,
        transpose=True,
        backend=implementation,
    )


def _csr_vector_api(x, data, indices, indptr, shape, implementation):
    return binary_csrmv(
        data,
        indices,
        indptr,
        x,
        shape=shape,
        transpose=False,
        backend=implementation,
    )


def _matrix_csr_api(x, data, indices, indptr, shape, implementation):
    # x @ csr: csrmm expects input as [shape[0], cols] for transpose=True.
    return binary_csrmm(
        data,
        indices,
        indptr,
        x.T,
        shape=shape,
        transpose=True,
        backend=implementation,
    ).T


def _csr_matrix_api(x, data, indices, indptr, shape, implementation):
    return binary_csrmm(
        data,
        indices,
        indptr,
        x,
        shape=shape,
        transpose=False,
        backend=implementation,
    )


class TestVectorCSR:
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_csr(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        m, n = 20, 40
        x = brainstate.random.rand(m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        y2 = vector_csr(x, data, indices, indptr, (m, n))

        for implementation in CSRMV_IMPLEMENTATIONS:
            y = _vector_csr_api(x, data, indices, indptr, (m, n), implementation)
            assert jnp.allclose(y, y2, rtol=1e-5, atol=1e-5)

        jax.block_until_ready((x, indptr, indices, y2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_csr_vmap_vector(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        with jax.checking_leaks():
            n_batch, m, n = 10, 20, 40
            xs = brainstate.random.rand(n_batch, m) < 0.1
            indptr, indices = get_csr(m, n, 0.1)

            data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
            y2 = brainstate.transform.vmap2(lambda x: vector_csr(x, data, indices, indptr, (m, n)))(xs)

            for implementation in CSRMV_IMPLEMENTATIONS:
                y = brainstate.transform.vmap2(
                    lambda x: _vector_csr_api(x, data, indices, indptr, (m, n), implementation)
                )(xs)
                assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((xs, indptr, indices))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_csr_vector(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        m, n = 20, 40
        v = brainstate.random.rand(n) < 0.1
        indptr, indices = get_csr(m, n, 0.2)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        y2 = csr_vector(v, data, indices, indptr, (m, n))

        for implementation in CSRMV_IMPLEMENTATIONS:
            y = _csr_vector_api(v, data, indices, indptr, (m, n), implementation)
            assert jnp.allclose(y, y2, rtol=1e-5, atol=1e-5)

        jax.block_until_ready((v, indptr, indices, y2))

    def _test_vjp(self, implementation, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in) if transpose else brainstate.random.rand(n_out)
        x = (x < 0.6).astype(float)

        indptr, indices = get_csr(n_in, n_out, 0.2, replace=replace)
        w = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)

        def f_api(x, w):
            if transpose:
                r = _vector_csr_api(x, w, indices, indptr, shape, implementation)
            else:
                r = _csr_vector_api(x, w, indices, indptr, shape, implementation)
            return r.sum()

        r = jax.grad(f_api, argnums=(0, 1))(x, w)

        def f_jax(x, w):
            if transpose:
                r = vector_csr(x, w, indices, indptr, shape=shape)
            else:
                r = csr_vector(x, w, indices, indptr, shape=shape)
            return r.sum()

        r2 = jax.grad(f_jax, argnums=(0, 1))(x, w)
        assert jnp.allclose(r[0], r2[0], rtol=1e-3, atol=1e-3)
        assert jnp.allclose(r[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_vjp(self, homo_w, replace, transpose):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._test_vjp(
                implementation=implementation,
                homo_w=homo_w,
                replace=replace,
                transpose=transpose,
            )

    def _test_jvp(self, implementation, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in if transpose else n_out)
        x = (x < 0.6).astype(float)

        indptr, indices = get_csr(n_in, n_out, 0.1, replace=replace)

        w = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)

        def f_api(x, w):
            if transpose:
                r = _vector_csr_api(x, w, indices, indptr, shape, implementation)
            else:
                r = _csr_vector_api(x, w, indices, indptr, shape, implementation)
            return r

        o1, r1 = jax.jvp(f_api, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))

        def f_jax(x, w):
            if transpose:
                r = vector_csr(x, w, indices, indptr, shape=shape)
            else:
                r = csr_vector(x, w, indices, indptr, shape=shape)
            return r

        o2, r2 = jax.jvp(f_jax, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        assert jnp.allclose(r1, r2, rtol=1e-3, atol=1e-3)
        assert jnp.allclose(o1, o2, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, o1, r1, o2, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_jvp(self, homo_w, replace, transpose):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._test_jvp(
                implementation=implementation,
                homo_w=homo_w,
                replace=replace,
                transpose=transpose,
            )


class TestBatchingVectorCSR:
    def _run(self, x, data, indices, indptr, m: int, n: int, transpose: bool = True):
        implementation = self._implementation
        if transpose:
            y1 = _vector_csr_api(x, data, indices, indptr, (m, n), implementation)
            y2 = vector_csr(x, data, indices, indptr, (m, n))
        else:
            y1 = _csr_vector_api(x, data, indices, indptr, (m, n), implementation)
            y2 = csr_vector(x, data, indices, indptr, (m, n))
        return jnp.allclose(y1, y2)

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_vector(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        xs = brainstate.random.rand(b, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            res = brainstate.transform.vmap2(lambda x: self._run(x, data, indices, indptr, m, n))(xs)
            assert jnp.all(res)

        jax.block_until_ready((xs, indptr, indices))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_data(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        x = brainstate.random.rand(m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = brainstate.random.rand(b) if homo_w else braintools.init.Normal(0., 1.)((b,) + indices.shape)
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            res = brainstate.transform.vmap2(lambda data: self._run(x, data, indices, indptr, m, n))(data)
            assert jnp.all(res)

        jax.block_until_ready((x, indptr, indices))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_indices(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        x = brainstate.random.rand(m) < 0.1
        indptr, indices = brainstate.transform.for_loop(lambda *a: get_csr(m, n, 0.1), length=b)
        indptr = indptr[0]

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape[1:])
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            res = brainstate.transform.vmap2(lambda ind: self._run(x, data, ind, indptr, m, n))(indices)
            assert jnp.all(res)

        jax.block_until_ready((x, indptr, indices))

    def _run_vjp(self, x, data, indices, indptr, m: int, n: int, transpose: bool = True):
        x = x.astype(float)
        implementation = self._implementation

        def f_api(x, w):
            if transpose:
                r = _vector_csr_api(x, w, indices, indptr, (m, n), implementation)
            else:
                r = _csr_vector_api(x, w, indices, indptr, (m, n), implementation)
            return r.sum()

        r1 = jax.grad(f_api, argnums=(0, 1))(x, data)

        def f_jax(x, w):
            if transpose:
                r = vector_csr(x, w, indices, indptr, shape=(m, n))
            else:
                r = csr_vector(x, w, indices, indptr, shape=(m, n))
            return r.sum()

        r2 = jax.grad(f_jax, argnums=(0, 1))(x, data)

        return r1, r2

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_vector_vjp(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        xs = brainstate.random.rand(b, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda x: self._run_vjp(x, data, indices, indptr, m, n))(xs)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((xs, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_data_vjp(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        x = brainstate.random.rand(m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = brainstate.random.rand(b) if homo_w else braintools.init.Normal(0., 1.)((b,) + indices.shape)
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda data: self._run_vjp(x, data, indices, indptr, m, n))(data)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_indices_vjp(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        x = brainstate.random.rand(m) < 0.1
        indptr, indices = brainstate.transform.for_loop(lambda *a: get_csr(m, n, 0.1), length=b)
        indptr = indptr[0]

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape[1:])
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda ind: self._run_vjp(x, data, ind, indptr, m, n))(indices)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))

    def _run_jvp(self, x, data, indices, indptr, m: int, n: int, transpose: bool = True):
        x = x.astype(float)
        implementation = self._implementation

        def f_api(x, w):
            if transpose:
                r = _vector_csr_api(x, w, indices, indptr, (m, n), implementation)
            else:
                r = _csr_vector_api(x, w, indices, indptr, (m, n), implementation)
            return r

        r1 = jax.jvp(f_api, (x, data), (jnp.ones_like(x), jnp.ones_like(data)))

        def f_jax(x, w):
            if transpose:
                r = vector_csr(x, w, indices, indptr, shape=(m, n))
            else:
                r = csr_vector(x, w, indices, indptr, shape=(m, n))
            return r

        r2 = jax.jvp(f_jax, (x, data), (jnp.ones_like(x), jnp.ones_like(data)))

        return r1, r2

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_vector_jvp(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        xs = brainstate.random.rand(b, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda x: self._run_jvp(x, data, indices, indptr, m, n))(xs)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((xs, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_data_jvp(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        x = brainstate.random.rand(m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = brainstate.random.rand(b) if homo_w else braintools.init.Normal(0., 1.)((b,) + indices.shape)
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda data: self._run_jvp(x, data, indices, indptr, m, n))(data)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_indices_jvp(self, homo_w):
        _require_implementations(CSRMV_IMPLEMENTATIONS, 'binary_csrmv')

        b, m, n = 10, 20, 40
        x = brainstate.random.rand(m) < 0.1
        indptr, indices = brainstate.transform.for_loop(lambda *a: get_csr(m, n, 0.1), length=b)
        indptr = indptr[0]

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape[1:])
        for implementation in CSRMV_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda ind: self._run_jvp(x, data, ind, indptr, m, n))(indices)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))


class TestMatrixCSR:
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_matrix_csr(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        k, m, n = 10, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        y2 = matrix_csr(x, data, indices, indptr, (m, n))

        for implementation in CSRMM_IMPLEMENTATIONS:
            y = _matrix_csr_api(x, data, indices, indptr, (m, n), implementation)
            assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, y2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_csr_matrix(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        m, n, k = 20, 40, 10
        matrix = brainstate.random.rand(n, k) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        y2 = csr_matrix(matrix, data, indices, indptr, (m, n))

        for implementation in CSRMM_IMPLEMENTATIONS:
            y = _csr_matrix_api(matrix, data, indices, indptr, (m, n), implementation)
            assert jnp.allclose(y, y2)

        jax.block_until_ready((matrix, indptr, indices, y2))


class TestBatchingMatrixCSR:
    def _run(self, x, data, indices, indptr, m: int, n: int, transpose: bool = True):
        implementation = self._implementation
        if transpose:
            y1 = _matrix_csr_api(x, data, indices, indptr, (m, n), implementation)
            y2 = matrix_csr(x, data, indices, indptr, (m, n))
        else:
            y1 = _csr_matrix_api(x, data, indices, indptr, (m, n), implementation)
            y2 = csr_matrix(x, data, indices, indptr, (m, n))
        return jnp.allclose(y1, y2)

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_matrix(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        xs = brainstate.random.rand(b, k, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            res = brainstate.transform.vmap2(lambda x: self._run(x, data, indices, indptr, m, n))(xs)
            assert jnp.all(res)

        jax.block_until_ready((xs, indptr, indices))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_data(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = brainstate.random.rand(b) if homo_w else braintools.init.Normal(0., 1.)((b,) + indices.shape)
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            res = brainstate.transform.vmap2(lambda data: self._run(x, data, indices, indptr, m, n))(data)
            assert jnp.all(res)

        jax.block_until_ready((x, indptr, indices))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_indices(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        indptr, indices = brainstate.transform.for_loop(lambda *a: get_csr(m, n, 0.1), length=b)
        indptr = indptr[0]

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape[1:])
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            res = brainstate.transform.vmap2(lambda ind: self._run(x, data, ind, indptr, m, n))(indices)
            assert jnp.all(res)

        jax.block_until_ready((x, indptr, indices))

    def _run_vjp(self, x, data, indices, indptr, m: int, n: int, transpose: bool = True):
        x = x.astype(float)
        implementation = self._implementation

        def f_api(x, w):
            if transpose:
                r = _matrix_csr_api(x, w, indices, indptr, (m, n), implementation)
            else:
                r = _csr_matrix_api(x, w, indices, indptr, (m, n), implementation)
            return r.sum()

        r1 = jax.grad(f_api, argnums=(0, 1))(x, data)

        def f_jax(x, w):
            if transpose:
                r = matrix_csr(x, w, indices, indptr, shape=(m, n))
            else:
                r = csr_matrix(x, w, indices, indptr, shape=(m, n))
            return r.sum()

        r2 = jax.grad(f_jax, argnums=(0, 1))(x, data)

        return r1, r2

    @pytest.mark.parametrize('transpose', [True, False])
    def test_vmap_matrix_vjp(self, transpose):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        if transpose:
            xs = brainstate.random.rand(b, n, m) < 0.1
        else:
            xs = brainstate.random.rand(b, n, k) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = braintools.init.Normal(0., 1.)(indices.shape)
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(
                lambda x: self._run_vjp(x, data, indices, indptr, m, n, transpose=transpose)
            )(xs)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((xs, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_data_vjp(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = brainstate.random.rand(b) if homo_w else braintools.init.Normal(0., 1.)((b,) + indices.shape)
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda data: self._run_vjp(x, data, indices, indptr, m, n))(data)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_indices_vjp(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        indptr, indices = brainstate.transform.for_loop(lambda *a: get_csr(m, n, 0.1), length=b)
        indptr = indptr[0]

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape[1:])
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda ind: self._run_vjp(x, data, ind, indptr, m, n))(indices)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))

    def _run_jvp(self, x, data, indices, indptr, m: int, n: int, transpose: bool = True):
        x = x.astype(float)
        implementation = self._implementation

        def f_api(x, w):
            if transpose:
                r = _matrix_csr_api(x, w, indices, indptr, (m, n), implementation)
            else:
                r = _csr_matrix_api(x, w, indices, indptr, (m, n), implementation)
            return r

        r1 = jax.jvp(f_api, (x, data), (jnp.ones_like(x), jnp.ones_like(data)))

        def f_jax(x, w):
            if transpose:
                r = matrix_csr(x, w, indices, indptr, shape=(m, n))
            else:
                r = csr_matrix(x, w, indices, indptr, shape=(m, n))
            return r

        r2 = jax.jvp(f_jax, (x, data), (jnp.ones_like(x), jnp.ones_like(data)))

        return r1, r2

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_vector_jvp(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        xs = brainstate.random.rand(b, k, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape)
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda x: self._run_jvp(x, data, indices, indptr, m, n))(xs)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((xs, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_data_jvp(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        indptr, indices = get_csr(m, n, 0.1)

        data = brainstate.random.rand(b) if homo_w else braintools.init.Normal(0., 1.)((b,) + indices.shape)
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda data: self._run_jvp(x, data, indices, indptr, m, n))(data)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vmap_indices_jvp(self, homo_w):
        _require_implementations(CSRMM_IMPLEMENTATIONS, 'binary_csrmm')

        b, k, m, n = 10, 15, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        indptr, indices = brainstate.transform.for_loop(lambda *a: get_csr(m, n, 0.1), length=b)
        indptr = indptr[0]

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(indices.shape[1:])
        for implementation in CSRMM_IMPLEMENTATIONS:
            self._implementation = implementation
            r1, r2 = brainstate.transform.vmap2(lambda ind: self._run_jvp(x, data, ind, indptr, m, n))(indices)
            assert jnp.allclose(r1[0], r2[0], rtol=1e-3, atol=1e-3)
            assert jnp.allclose(r1[1], r2[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, r1, r2))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-

from typing import Optional

from jax.tree_util import register_pytree_node_class

from .base import IndexedEventRepresentation, is_known_type
from .binary_indexed_extraction import binary_1d_array_index_p_call, binary_2d_array_index_p_call

__all__ = [
    'IndexedBinary1d',
    'IndexedBinary2d',
]


@register_pytree_node_class
class IndexedBinary1d(IndexedEventRepresentation):
    """A 1-D binary event array with pre-computed indices of active elements.

    On construction the input array is scanned to extract the positions of
    all non-zero (active) elements, stored as ``spike_indices``, together
    with the total count ``spike_count``.  This pre-indexing enables
    downstream kernels to iterate only over active elements.

    Parameters
    ----------
    value : array_like
        A 1-D boolean or numeric array.  Non-zero entries are considered
        active events.
    backend : str or None, optional
        Compute backend for the index-extraction kernel (e.g. ``'numba'``,
        ``'warp'``, ``'pallas'``).  ``None`` selects the default.

    Notes
    -----
    The index extraction scans ``value`` and collects positions where
    ``value[i] != 0`` into a dense index array of shape ``(k,)`` where
    ``k`` equals the length of ``value``.  Unused trailing entries are
    zero-filled.  The actual number of active indices is given by
    ``spike_count[0]``.

    See Also
    --------
    IndexedBinary2d : 2-D variant with per-row index extraction.
    BinaryArray : Simpler wrapper without pre-computed indices.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> import brainevent as be
        >>> from brainevent._event.binary_indexed import IndexedBinary1d
        >>> spikes = IndexedBinary1d(jnp.array([True, False, True, False]))
        >>> spikes.spike_count
        Array([2], dtype=int32)
    """
    __module__ = 'brainevent'

    def __init__(self, value, backend: Optional[str] = None):
        super().__init__(value)
        self._spike_indices, self._spike_count = binary_1d_array_index_p_call(self._value, backend=backend)

    @property
    def spike_indices(self):
        """Indices of non-zero elements in the array.

        Returns
        -------
        jax.Array
            An int32 array of shape ``(k,)`` (where ``k == len(value)``)
            containing the positions of active elements in the first
            ``spike_count[0]`` entries.  Trailing entries are zero-filled.
        """
        return self._spike_indices

    @property
    def spike_count(self):
        """Number of active (non-zero) elements.

        Returns
        -------
        jax.Array
            An int32 array of shape ``(1,)`` holding the count of non-zero
            entries.
        """
        return self._spike_count

    def __matmul__(self, oc):
        """Matrix multiplication is not supported for ``IndexedBinary1d``.

        Raises
        ------
        ValueError
            Always raised — use ``BinaryArray`` for ``@`` operations.
        """
        raise ValueError

    def __rmatmul__(self, oc):
        """Reverse matrix multiplication is not supported for ``IndexedBinary1d``.

        Raises
        ------
        ValueError
            Always raised — use ``BinaryArray`` for ``@`` operations.
        """
        raise ValueError

    def __imatmul__(self, oc):
        """In-place matrix multiplication (not truly in-place under JAX).

        Parameters
        ----------
        oc : array_like
            Right operand.

        Returns
        -------
        IndexedBinary1d
            A new instance wrapping the result.

        Raises
        ------
        ValueError
            Delegates to ``__matmul__`` which always raises.
        """
        if is_known_type(oc):
            return self.with_value(self.__matmul__(oc))
        return self.with_value(oc.__rmatmul__(self))

    def tree_flatten(self):
        """Flatten for JAX PyTree serialisation.

        Returns
        -------
        children : tuple
            ``(value,)`` — the dynamic array leaf.
        aux_data : dict
            Contains ``_spike_indices`` and ``_spike_count`` as static
            metadata.
        """
        aux = {
            '_spike_indices': self._spike_indices,
            '_spike_count': self._spike_count,
        }
        return (self._value,), aux

    @classmethod
    def tree_unflatten(cls, aux_data, flat_contents):
        """Reconstruct from PyTree representation.

        Parameters
        ----------
        aux_data : dict
            Static metadata including spike indices and count.
        flat_contents : tuple
            The dynamic array leaf.

        Returns
        -------
        IndexedBinary1d
            Reconstructed instance.
        """
        value, = flat_contents
        obj = object.__new__(cls)
        obj._value = value
        for k, v in aux_data.items():
            setattr(obj, k, v)
        return obj


@register_pytree_node_class
class IndexedBinary2d(IndexedEventRepresentation):
    """A 2-D binary event array with pre-computed indices of active elements.

    Extends ``IndexedEventRepresentation`` for 2-D binary arrays, extracting
    per-row indices of non-zero elements on construction.

    Parameters
    ----------
    value : array_like
        A 2-D boolean or numeric array of shape ``(batch, k)``.
    row_indices : None, optional
        Reserved for future use.

    Notes
    -----
    The 2-D index extraction is not yet implemented and will raise
    ``NotImplementedError`` during construction.

    See Also
    --------
    IndexedBinary1d : 1-D variant.
    """
    __module__ = 'brainevent'

    def __init__(
        self,
        value,
        *,
        row_indices=None,
    ):
        super().__init__(value)
        self._spike_indices, self._spike_count = binary_2d_array_index_p_call(self._value)

    @property
    def spike_indices(self):
        """Indices of non-zero elements per row.

        Returns
        -------
        jax.Array
            Index array for the active elements.
        """
        return self._spike_indices

    @property
    def spike_count(self):
        """Number of active elements per row.

        Returns
        -------
        jax.Array
            Count array for non-zero entries.
        """
        return self._spike_count

    def __matmul__(self, oc):
        """Matrix multiplication is not supported for ``IndexedBinary2d``.

        Raises
        ------
        ValueError
            Always raised.
        """
        raise ValueError

    def __rmatmul__(self, oc):
        """Reverse matrix multiplication is not supported for ``IndexedBinary2d``.

        Raises
        ------
        ValueError
            Always raised.
        """
        raise ValueError

    def __imatmul__(self, oc):
        """In-place matrix multiplication (not truly in-place under JAX).

        Parameters
        ----------
        oc : array_like
            Right operand.

        Returns
        -------
        IndexedBinary2d
            A new instance wrapping the result.

        Raises
        ------
        ValueError
            Delegates to ``__matmul__`` which always raises.
        """
        if is_known_type(oc):
            return self.with_value(self.__matmul__(oc))
        return self.with_value(oc.__rmatmul__(self))

    def tree_flatten(self):
        """Flatten for JAX PyTree serialisation.

        Returns
        -------
        children : tuple
            ``(value,)`` — the dynamic array leaf.
        aux_data : dict
            Contains ``_spike_indices`` and ``_spike_count``.
        """
        aux = {
            '_spike_indices': self._spike_indices,
            '_spike_count': self._spike_count,
        }
        return (self._value,), aux

    @classmethod
    def tree_unflatten(cls, aux_data, flat_contents):
        """Reconstruct from PyTree representation.

        Parameters
        ----------
        aux_data : dict
            Static metadata including spike indices and count.
        flat_contents : tuple
            The dynamic array leaf.

        Returns
        -------
        IndexedBinary2d
            Reconstructed instance.
        """
        value, = flat_contents
        obj = object.__new__(cls)
        obj._value = value
        for k, v in aux_data.items():
            setattr(obj, k, v)
        return obj



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from brainevent._data import _initialize_conn_length, _initialize_seed
from brainevent._jit_normal.binary import (
    binary_jitnmv,
    binary_jitnmv_p,
    binary_jitnmm_p,
    binary_jitnmm_p_call,
)
from brainevent._jit_normal.float import jitnmv, jitnmm
from brainevent._test_util import allclose

platform = jax.default_backend()
JITNMV_IMPLEMENTATIONS = tuple(binary_jitnmv_p.available_backends(platform))
JITNMM_IMPLEMENTATIONS = tuple(binary_jitnmm_p.available_backends(platform))

if platform == 'cpu':
    SHAPES = ((20, 30), (100, 50))
else:
    SHAPES = ((20, 30), (100, 50), (400, 300))


def _implementation_params(implementations, op_name: str):
    if implementations:
        return [pytest.param(impl, id=impl) for impl in implementations]
    return [
        pytest.param(
            None,
            marks=pytest.mark.skip(reason=f'No {op_name} implementations on platform={platform}'),
            id=f'no-{op_name}',
        )
    ]


JITNMV_PARAMS = _implementation_params(JITNMV_IMPLEMENTATIONS, 'binary_jitnmv')
JITNMM_PARAMS = _implementation_params(JITNMM_IMPLEMENTATIONS, 'binary_jitnmm')


def _binary_events(x, dtype=jnp.float32):
    return jnp.asarray(jnp.asarray(x) > 0, dtype=dtype)


def _sample_vector(size: int, event_dtype, seed: int):
    rng = np.random.RandomState(seed)
    if event_dtype is bool:
        return jnp.asarray(rng.rand(size) > 0.5)
    raw = rng.rand(size).astype(np.float32)
    mask = rng.rand(size) > 0.5
    return jnp.asarray(raw * mask)


def _sample_matrix(rows: int, cols: int, event_dtype, seed: int):
    rng = np.random.RandomState(seed)
    if event_dtype is bool:
        return jnp.asarray(rng.rand(rows, cols) > 0.5)
    raw = rng.rand(rows, cols).astype(np.float32)
    mask = rng.rand(rows, cols) > 0.5
    return jnp.asarray(raw * mask)


def _call_binary_jitnmm(w_loc, w_scale, prob, matrix, seed, *, shape, transpose, corder, implementation):
    return binary_jitnmm_p_call(
        w_loc,
        w_scale,
        _initialize_conn_length(prob),
        matrix,
        _initialize_seed(seed),
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=implementation,
    )[0]


@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('shape', SHAPES)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('event_dtype', [bool, float])
@pytest.mark.parametrize('prob', [0.1, 0.2])
def test_binary_jitnmv_forward_matches_reference(implementation, shape, transpose, corder, event_dtype, prob):
    seed = 123
    w_loc = jnp.asarray(1.5, dtype=jnp.float32)
    w_scale = jnp.asarray(0.1, dtype=jnp.float32)
    event_size = shape[0] if transpose else shape[1]
    vector = _sample_vector(event_size, event_dtype, seed + 7)
    vector_ref = _binary_events(vector, dtype=jnp.float32)

    y = binary_jitnmv(
        w_loc,
        w_scale,
        prob,
        vector,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=implementation,
    )
    y_ref = jitnmv(
        w_loc,
        w_scale,
        prob,
        vector_ref,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=implementation,
    )
    assert allclose(y, y_ref, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((w_loc, w_scale, vector, vector_ref, y, y_ref))


@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('shape', SHAPES)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('event_dtype', [bool, float])
@pytest.mark.parametrize('prob', [0.1, 0.2])
@pytest.mark.parametrize('k', [5, 10])
def test_binary_jitnmm_forward_matches_reference(implementation, shape, transpose, corder, event_dtype, prob, k):
    seed = 123
    w_loc = jnp.asarray([1.5], dtype=jnp.float32)
    w_scale = jnp.asarray([0.1], dtype=jnp.float32)
    rows = shape[0] if transpose else shape[1]
    matrix = _sample_matrix(rows, k, event_dtype, seed + 11)
    matrix_ref = _binary_events(matrix, dtype=jnp.float32)

    y = _call_binary_jitnmm(
        w_loc,
        w_scale,
        prob,
        matrix,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        implementation=implementation,
    )
    y_ref = jitnmm(
        w_loc[0],
        w_scale[0],
        prob,
        matrix_ref,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=implementation,
    )
    assert allclose(y, y_ref, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((w_loc, w_scale, matrix, matrix_ref, y, y_ref))


@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
def test_binary_jitnmv_thresholds_float_events(implementation, shape, transpose, corder):
    seed = 123
    prob = 0.1
    w_loc = jnp.asarray(1.5, dtype=jnp.float32)
    w_scale = jnp.asarray(0.1, dtype=jnp.float32)
    size = shape[0] if transpose else shape[1]
    vector = _sample_vector(size, float, seed + 17)
    vector_binary = _binary_events(vector, dtype=jnp.float32)

    y_float = binary_jitnmv(
        w_loc,
        w_scale,
        prob,
        vector,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=implementation,
    )
    y_binary = binary_jitnmv(
        w_loc,
        w_scale,
        prob,
        vector_binary,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=implementation,
    )
    assert allclose(y_float, y_binary, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((w_loc, w_scale, vector, vector_binary, y_float, y_binary))


@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('k', [10])
def test_binary_jitnmm_thresholds_float_events(implementation, shape, transpose, corder, k):
    seed = 123
    prob = 0.1
    w_loc = jnp.asarray([1.5], dtype=jnp.float32)
    w_scale = jnp.asarray([0.1], dtype=jnp.float32)
    rows = shape[0] if transpose else shape[1]
    matrix = _sample_matrix(rows, k, float, seed + 23)
    matrix_binary = _binary_events(matrix, dtype=jnp.float32)

    y_float = _call_binary_jitnmm(
        w_loc,
        w_scale,
        prob,
        matrix,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        implementation=implementation,
    )
    y_binary = _call_binary_jitnmm(
        w_loc,
        w_scale,
        prob,
        matrix_binary,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        implementation=implementation,
    )
    assert allclose(y_float, y_binary, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((w_loc, w_scale, matrix, matrix_binary, y_float, y_binary))


@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
def test_binary_jitnmv_jvp_and_vjp_match_reference(implementation, transpose, corder):
    shape = (20, 30)
    seed = 123
    prob = 0.1
    vector_size = shape[0] if transpose else shape[1]
    vector = _binary_events(_sample_vector(vector_size, float, seed + 29), dtype=jnp.float32)

    def f_binary(w_loc, w_scale, v):
        return binary_jitnmv(
            w_loc,
            w_scale,
            prob,
            v,
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            backend=implementation,
        )

    def f_ref(w_loc, w_scale, v):
        return jitnmv(
            w_loc,
            w_scale,
            prob,
            v,
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            backend=implementation,
        )

    primals = (jnp.asarray(1.5, dtype=jnp.float32), jnp.asarray(0.1, dtype=jnp.float32), vector)
    tangents = (jnp.asarray(1.0, dtype=jnp.float32), jnp.asarray(0.5, dtype=jnp.float32), jnp.ones_like(vector))

    out1, jvp1 = jax.jvp(f_binary, primals, tangents)
    out2, jvp2 = jax.jvp(f_ref, primals, tangents)
    assert allclose(out1, out2, rtol=1e-4, atol=1e-4)
    assert allclose(jvp1, jvp2, rtol=1e-4, atol=1e-4)

    def scalar_binary(w_loc, w_scale, v):
        return f_binary(w_loc, w_scale, v).sum()

    def scalar_ref(w_loc, w_scale, v):
        return f_ref(w_loc, w_scale, v).sum()

    (g_wloc1, g_wscale1, g_v1) = jax.grad(scalar_binary, argnums=(0, 1, 2))(*primals)
    (g_wloc2, g_wscale2, g_v2) = jax.grad(scalar_ref, argnums=(0, 1, 2))(*primals)
    assert allclose(g_wloc1, g_wloc2, rtol=1e-4, atol=1e-4)
    assert allclose(g_wscale1, g_wscale2, rtol=1e-4, atol=1e-4)
    assert allclose(g_v1, g_v2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vector, primals[0], primals[1], tangents[0], tangents[1], tangents[2],
                           out1, jvp1, out2, jvp2, g_wloc1, g_wscale1, g_v1, g_wloc2, g_wscale2, g_v2))


@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
def test_binary_jitnmm_jvp_matches_reference(implementation, transpose, corder):
    shape = (20, 30)
    seed = 123
    prob = 0.1
    k = 8
    rows = shape[0] if transpose else shape[1]
    matrix = _binary_events(_sample_matrix(rows, k, float, seed + 31), dtype=jnp.float32)

    def f_binary(w_loc, w_scale, B):
        return _call_binary_jitnmm(
            jnp.atleast_1d(w_loc),
            jnp.atleast_1d(w_scale),
            prob,
            B,
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            implementation=implementation,
        )

    def f_ref(w_loc, w_scale, B):
        return jitnmm(
            w_loc,
            w_scale,
            prob,
            B,
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            backend=implementation,
        )

    primals = (jnp.asarray(1.5, dtype=jnp.float32), jnp.asarray(0.1, dtype=jnp.float32), matrix)
    tangents = (jnp.asarray(1.0, dtype=jnp.float32), jnp.asarray(0.5, dtype=jnp.float32), jnp.ones_like(matrix))
    out1, jvp1 = jax.jvp(f_binary, primals, tangents)
    out2, jvp2 = jax.jvp(f_ref, primals, tangents)
    assert allclose(out1, out2, rtol=1e-4, atol=1e-4)
    assert allclose(jvp1, jvp2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((matrix, primals[0], primals[1], tangents[0], tangents[1], tangents[2],
                           out1, jvp1, out2, jvp2))


@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
def test_binary_jitnmv_vmap_matches_reference(implementation, transpose, corder):
    shape = (20, 30)
    seed = 123
    prob = 0.1
    batch = 6
    event_size = shape[0] if transpose else shape[1]
    vectors = _sample_matrix(batch, event_size, float, seed + 41)

    f_binary = jax.vmap(
        lambda v: binary_jitnmv(
            jnp.asarray(1.5, dtype=jnp.float32),
            jnp.asarray(0.1, dtype=jnp.float32),
            prob,
            v,
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            backend=implementation,
        )
    )
    f_ref = jax.vmap(
        lambda v: jitnmv(
            jnp.asarray(1.5, dtype=jnp.float32),
            jnp.asarray(0.1, dtype=jnp.float32),
            prob,
            _binary_events(v, dtype=jnp.float32),
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            backend=implementation,
        )
    )
    result_binary = f_binary(vectors)
    result_ref = f_ref(vectors)
    assert allclose(result_binary, result_ref, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vectors, result_binary, result_ref))


@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
def test_binary_jitnmm_vmap_matches_reference(implementation, transpose, corder):
    shape = (20, 30)
    seed = 123
    prob = 0.1
    batch = 4
    k = 7
    rows = shape[0] if transpose else shape[1]
    matrices = _sample_matrix(batch * rows, k, float, seed + 43).reshape(batch, rows, k)

    f_binary = jax.vmap(
        lambda B: _call_binary_jitnmm(
            jnp.asarray([1.5], dtype=jnp.float32),
            jnp.asarray([0.1], dtype=jnp.float32),
            prob,
            B,
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            implementation=implementation,
        )
    )
    f_ref = jax.vmap(
        lambda B: jitnmm(
            jnp.asarray(1.5, dtype=jnp.float32),
            jnp.asarray(0.1, dtype=jnp.float32),
            prob,
            _binary_events(B, dtype=jnp.float32),
            seed,
            shape=shape,
            transpose=transpose,
            corder=corder,
            backend=implementation,
        )
    )
    result_binary = f_binary(matrices)
    result_ref = f_ref(matrices)
    assert allclose(result_binary, result_ref, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((matrices, result_binary, result_ref))


# ---- Gradient VJP: binary_jitnmv w.r.t. w_loc ----

@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmv_vjp_wloc(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    vec_size = shape[0] if transpose else shape[1]
    vector = _binary_events(_sample_vector(vec_size, float, seed + 7), dtype=jnp.float32)
    w_loc_arr = jnp.array([w_loc])

    def f_fn(wl):
        return binary_jitnmv(wl, w_scale, prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                             backend=implementation).sum()

    def f_ref(wl):
        return jitnmv(wl[0], w_scale, prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    grad1 = jax.grad(f_fn)(w_loc_arr)
    grad2 = jax.grad(f_ref)(w_loc_arr)
    assert allclose(grad1, grad2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vector, w_loc_arr, grad1, grad2))


# ---- Gradient VJP: binary_jitnmv w.r.t. w_scale ----

@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmv_vjp_wscale(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    vec_size = shape[0] if transpose else shape[1]
    vector = _binary_events(_sample_vector(vec_size, float, seed + 7), dtype=jnp.float32)
    w_scale_arr = jnp.array([w_scale])

    def f_fn(ws):
        return binary_jitnmv(w_loc, ws, prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                             backend=implementation).sum()

    def f_ref(ws):
        return jitnmv(w_loc, ws[0], prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    grad1 = jax.grad(f_fn)(w_scale_arr)
    grad2 = jax.grad(f_ref)(w_scale_arr)
    assert allclose(grad1, grad2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vector, w_scale_arr, grad1, grad2))


# ---- End-to-end VJP: binary_jitnmv w.r.t. w_loc with loss ----

@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmv_vjp_wloc_with_loss(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    vec_size = shape[0] if transpose else shape[1]
    out_size = shape[1] if transpose else shape[0]
    vector = _binary_events(_sample_vector(vec_size, float, seed + 7), dtype=jnp.float32)
    target = jnp.asarray(np.random.rand(out_size))
    w_loc_arr = jnp.array([w_loc])

    def loss_fn(wl):
        out = binary_jitnmv(wl, w_scale, prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                            backend=implementation)
        return jnp.sum((out - target) ** 2)

    def loss_ref(wl):
        out = jitnmv(wl[0], w_scale, prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                     backend=implementation)
        return jnp.sum((out - target) ** 2)

    grad1 = jax.grad(loss_fn)(w_loc_arr)
    grad2 = jax.grad(loss_ref)(w_loc_arr)
    assert allclose(grad1, grad2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vector, target, w_loc_arr, grad1, grad2))


# ---- End-to-end VJP: binary_jitnmv w.r.t. w_scale with loss ----

@pytest.mark.parametrize('implementation', JITNMV_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmv_vjp_wscale_with_loss(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    vec_size = shape[0] if transpose else shape[1]
    out_size = shape[1] if transpose else shape[0]
    vector = _binary_events(_sample_vector(vec_size, float, seed + 7), dtype=jnp.float32)
    target = jnp.asarray(np.random.rand(out_size))
    w_scale_arr = jnp.array([w_scale])

    def loss_fn(ws):
        out = binary_jitnmv(w_loc, ws, prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                            backend=implementation)
        return jnp.sum((out - target) ** 2)

    def loss_ref(ws):
        out = jitnmv(w_loc, ws[0], prob, vector, seed, shape=shape, transpose=transpose, corder=corder,
                     backend=implementation)
        return jnp.sum((out - target) ** 2)

    grad1 = jax.grad(loss_fn)(w_scale_arr)
    grad2 = jax.grad(loss_ref)(w_scale_arr)
    assert allclose(grad1, grad2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vector, target, w_scale_arr, grad1, grad2))


# ---- Gradient VJP: binary_jitnmm w.r.t. w_loc ----

@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmm_vjp_wloc(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    k = 10
    mat_rows = shape[0] if transpose else shape[1]
    B = _binary_events(_sample_matrix(mat_rows, k, float, seed + 31), dtype=jnp.float32)
    w_loc_arr = jnp.array([w_loc])

    def f_fn(wl):
        return _call_binary_jitnmm(wl, jnp.array([w_scale]), prob, B, seed,
                                   shape=shape, transpose=transpose, corder=corder, implementation=implementation).sum()

    def f_ref(wl):
        return jitnmm(wl[0], w_scale, prob, B, seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    grad1 = jax.grad(f_fn)(w_loc_arr)
    grad2 = jax.grad(f_ref)(w_loc_arr)
    assert allclose(grad1, grad2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((B, w_loc_arr, grad1, grad2))


# ---- Gradient VJP: binary_jitnmm w.r.t. w_scale ----

@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmm_vjp_wscale(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    k = 10
    mat_rows = shape[0] if transpose else shape[1]
    B = _binary_events(_sample_matrix(mat_rows, k, float, seed + 31), dtype=jnp.float32)
    w_scale_arr = jnp.array([w_scale])

    def f_fn(ws):
        return _call_binary_jitnmm(jnp.array([w_loc]), ws, prob, B, seed,
                                   shape=shape, transpose=transpose, corder=corder, implementation=implementation).sum()

    def f_ref(ws):
        return jitnmm(w_loc, ws[0], prob, B, seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    grad1 = jax.grad(f_fn)(w_scale_arr)
    grad2 = jax.grad(f_ref)(w_scale_arr)
    assert allclose(grad1, grad2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((B, w_scale_arr, grad1, grad2))


# ---- End-to-end VJP: binary_jitnmm w.r.t. w_loc with loss ----

@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmm_vjp_wloc_with_loss(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    k = 10
    mat_rows = shape[0] if transpose else shape[1]
    out_rows = shape[1] if transpose else shape[0]
    B = _binary_events(_sample_matrix(mat_rows, k, float, seed + 31), dtype=jnp.float32)
    target = jnp.asarray(np.random.rand(out_rows, k))
    w_loc_arr = jnp.array([w_loc])

    def loss_fn(wl):
        out = _call_binary_jitnmm(wl, jnp.array([w_scale]), prob, B, seed,
                                  shape=shape, transpose=transpose, corder=corder, implementation=implementation)
        return jnp.sum((out - target) ** 2)

    def loss_ref(wl):
        out = jitnmm(wl[0], w_scale, prob, B, seed, shape=shape, transpose=transpose, corder=corder,
                     backend=implementation)
        return jnp.sum((out - target) ** 2)

    grad1 = jax.grad(loss_fn)(w_loc_arr)
    grad2 = jax.grad(loss_ref)(w_loc_arr)
    assert allclose(grad1, grad2, rtol=2e-2, atol=2e-2)
    jax.block_until_ready((B, target, w_loc_arr, grad1, grad2))


# ---- End-to-end VJP: binary_jitnmm w.r.t. w_scale with loss ----

@pytest.mark.parametrize('implementation', JITNMM_PARAMS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_binary_jitnmm_vjp_wscale_with_loss(implementation, shape, corder, transpose):
    w_loc, w_scale, prob, seed = 1.5, 0.15, 0.1, 123
    k = 10
    mat_rows = shape[0] if transpose else shape[1]
    out_rows = shape[1] if transpose else shape[0]
    B = _binary_events(_sample_matrix(mat_rows, k, float, seed + 31), dtype=jnp.float32)
    target = jnp.asarray(np.random.rand(out_rows, k))
    w_scale_arr = jnp.array([w_scale])

    def loss_fn(ws):
        out = _call_binary_jitnmm(jnp.array([w_loc]), ws, prob, B, seed,
                                  shape=shape, transpose=transpose, corder=corder, implementation=implementation)
        return jnp.sum((out - target) ** 2)

    def loss_ref(ws):
        out = jitnmm(w_loc, ws[0], prob, B, seed, shape=shape, transpose=transpose, corder=corder,
                     backend=implementation)
        return jnp.sum((out - target) ** 2)

    grad1 = jax.grad(loss_fn)(w_scale_arr)
    grad2 = jax.grad(loss_ref)(w_scale_arr)
    assert allclose(grad1, grad2, rtol=2e-2, atol=2e-2)
    jax.block_until_ready((B, target, w_scale_arr, grad1, grad2))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-

from typing import Optional

import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np
from jax.interpreters import ad

from brainevent._misc import cdiv, generate_block_dim, namescope
from brainevent._op import jaxinfo_to_warpinfo, numba_kernel, XLACustomKernel, general_batching_rule
from brainevent._op.benchmark import BenchmarkConfig
from brainevent.config import get_numba_parallel

__all__ = [
    'binary_densemv', 'binary_densemv_p', 'binary_densemv_p_call',
    'binary_densemm', 'binary_densemm_p', 'binary_densemm_p_call',
]


# ==============================================================================
# Unified binary dense matrix-vector product (binary_densemv)
# ==============================================================================
#
# transpose=False: weights[m,k] @ spikes[k] -> out[m]  (old dbmv)
# transpose=True:  spikes[k] @ weights[k,n] -> out[n]  (old bdvm)
#
# Argument order is always (weights, spikes).


@namescope(static_argnames=['transpose'])
def binary_densemv(weights, spikes, *, transpose, backend: Optional[str] = None):
    """
    Perform event-driven dense matrix-vector multiplication with binary spikes.

    When ``transpose=False``, computes ``weights[m, k] @ spikes[k] -> out[m]``
    (dense matrix times binary vector).

    When ``transpose=True``, computes ``spikes[k] @ weights[k, n] -> out[n]``
    (binary vector times dense matrix).

    Parameters
    ----------
    weights : array_like
        The weight matrix. Shape ``(m, k)`` when ``transpose=False``,
        or ``(k, n)`` when ``transpose=True``. Can be a ``brainunit`` quantity.
    spikes : array_like
        The binary vector with shape ``(k,)``. Can be boolean or float.
        Can be a ``brainunit`` quantity.
    transpose : bool
        If False, compute ``weights @ spikes``. If True, compute
        ``spikes @ weights``.
    backend : str, optional
        Backend to use for the computation. One of ``'numba'``, ``'warp'``,
        ``'pallas'``, or ``None`` (auto-select).

    Returns
    -------
    result : array_like
        Result vector. Shape ``(m,)`` when ``transpose=False``,
        or ``(n,)`` when ``transpose=True``. If inputs carry units, the
        result carries the product of the weight and spike units.

    Raises
    ------
    AssertionError
        If the inner dimensions of ``weights`` and ``spikes`` do not match.

    See Also
    --------
    binary_densemm : Matrix-matrix variant of binary dense multiplication.

    Notes
    -----
    The computation is event-driven: only the columns (or rows) of
    ``weights`` corresponding to active (nonzero or True) entries of
    ``spikes`` are accumulated. This is mathematically equivalent to a
    standard matrix-vector product but can be faster when spikes are sparse.

    When ``transpose=False``, the operation computes:

    ``out[i] = sum_{j} W[i, j] * s[j]``

    where ``s[j]`` is ``1`` if ``spikes[j]`` is active and ``0`` otherwise.

    When ``transpose=True``, the operation computes:

    ``out[j] = sum_{i} W[i, j] * s[i]``

    where ``s[i]`` is ``1`` if ``spikes[i]`` is active and ``0`` otherwise.

    For boolean spikes, an entry is considered active when it is ``True``.
    For float spikes, an entry is considered active when it is ``> 0``.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._dense.binary import binary_densemv
        >>> weights = jnp.ones((3, 4), dtype=jnp.float32)
        >>> spikes = jnp.array([True, False, True, False])
        >>> binary_densemv(weights, spikes, transpose=False)
        Array([2., 2., 2.], dtype=float32)
    """
    with jax.ensure_compile_time_eval():
        weights = u.math.asarray(weights)
        spikes = u.math.asarray(spikes)
    weight_val, wunit = u.split_mantissa_unit(weights)
    spk_val, spkunit = u.split_mantissa_unit(spikes)
    r = binary_densemv_p_call(weight_val, spk_val, transpose=transpose, backend=backend)
    return u.maybe_decimal(r[0] * wunit * spkunit)


def _binary_densemv_numba_kernel(
    spk_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import numba

    if transpose:
        # weights[k,n], spikes[k] -> out[n]
        if spk_info.dtype == jnp.bool_:
            @numba.njit(fastmath=True)
            def kernel(weights, spikes, posts):
                posts[:] = 0.
                for i in range(spikes.shape[0]):
                    if spikes[i]:
                        posts += weights[i]
        else:
            @numba.njit(fastmath=True)
            def kernel(weights, spikes, posts):
                posts[:] = 0.
                for i in range(spikes.shape[0]):
                    if spikes[i] > 0.:
                        posts += weights[i]
    else:
        # weights[m,k], spikes[k] -> out[m]
        if spk_info.dtype == jnp.bool_:
            @numba.njit(fastmath=True)
            def kernel(weights, spikes, posts):
                posts[:] = 0.
                for i in range(spikes.shape[0]):
                    if spikes[i]:
                        posts += weights[:, i]
        else:
            @numba.njit(fastmath=True)
            def kernel(weights, spikes, posts):
                posts[:] = 0.
                for i in range(spikes.shape[0]):
                    if spikes[i] > 0.:
                        posts += weights[:, i]

    def run(weights, spikes):
        return numba_kernel(kernel, outs=kwargs['outs'])(weights, spikes)

    return run


def _binary_densemv_warp_kernel(
    weight_info: jax.ShapeDtypeStruct,
    spk_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    spike_length = spk_info.shape[0]
    weight_warp_info = jaxinfo_to_warpinfo(weight_info)
    spike_warp_info = jaxinfo_to_warpinfo(spk_info)
    out_warp_info = jaxinfo_to_warpinfo(kwargs['outs'][0])

    if transpose:
        # weights[k,n], spikes[k] -> out[n]
        n = weight_info.shape[1]

        if spk_info.dtype == jnp.bool_:
            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info,
            ):
                j = warp.tid()
                r = weights.dtype(0.)
                for i in range(spike_length):
                    if spikes[i]:
                        r += weights[i, j]
                out[j] = r
        else:
            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info,
            ):
                j = warp.tid()
                r = weights.dtype(0.)
                for i in range(spike_length):
                    if spikes[i] > 0.:
                        r += weights[i, j]
                out[j] = r

        def run(weights, spikes):
            out_info = kwargs['outs'][0]
            fn = jax_kernel(kernel, launch_dims=[n], num_outputs=1, output_dims={'out': out_info.shape})
            return fn(weights, spikes)

    else:
        # weights[m,k], spikes[k] -> out[m]
        m = weight_info.shape[0]

        if spk_info.dtype == jnp.bool_:
            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info,
            ):
                i = warp.tid()
                r = weights.dtype(0.)
                for j in range(spike_length):
                    if spikes[j]:
                        r += weights[i, j]
                out[i] = r
        else:
            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info,
            ):
                i = warp.tid()
                r = weights.dtype(0.)
                for j in range(spike_length):
                    if spikes[j] > 0.:
                        r += weights[i, j]
                out[i] = r

        def run(weights, spikes):
            out_info = kwargs['outs'][0]
            fn = jax_kernel(kernel, launch_dims=[m], num_outputs=1, output_dims={'out': out_info.shape})
            return fn(weights, spikes)

    return run


def _binary_densemv_pallas_kernel(
    weight_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    from jax.experimental import pallas as pl

    if transpose:
        # weights[k,n], spikes[k] -> out[n]
        n = weight_info.shape[1]
        block_dim = generate_block_dim(n, maximum=1024)

        def kernel(weight_ref, spike_ref, out_ref):
            i_col_block = pl.program_id(0)
            i_col_start = i_col_block * block_dim
            i_col_mask = i_col_start + jnp.arange(block_dim) < weight_ref.shape[1]

            def loop_fn(i_spike, temp):
                spike = spike_ref[i_spike]
                weight_row = weight_ref[i_spike, pl.ds(i_col_start, block_dim)]
                weight_row = jnp.where(i_col_mask, weight_row, 0.0)
                return jax.lax.cond(
                    spike if (spike_ref.dtype == jnp.bool_) else spike > 0.,
                    lambda out: out + weight_row,
                    lambda out: out,
                    temp,
                )

            i_col_out = jax.lax.fori_loop(
                0,
                spike_ref.shape[0],
                loop_fn,
                jnp.zeros((block_dim,), dtype=weight_ref.dtype)
            )
            out_ref[pl.ds(i_col_start, block_dim)] = jnp.where(i_col_mask, i_col_out, 0.0)

        def run(weights, spikes):
            fn = pl.pallas_call(kernel, grid=(cdiv(n, block_dim),), out_shape=kwargs['outs'], backend='triton')
            return fn(weights, spikes)

    else:
        # weights[m,k], spikes[k] -> out[m]
        m = weight_info.shape[0]
        mat_block_dim = generate_block_dim(m, maximum=1024)

        def kernel(weight_ref, spike_ref, out_ref):
            i_row_block = pl.program_id(0)
            i_row_start = i_row_block * mat_block_dim
            i_row_mask = i_row_start + jnp.arange(mat_block_dim) < weight_ref.shape[0]

            def loop_fn(i_spike, temp):
                spike = spike_ref[i_spike]
                weight_col = weight_ref[pl.ds(i_row_start, mat_block_dim), i_spike]
                weight_col = jnp.where(i_row_mask, weight_col, 0.0)
                return jax.lax.cond(
                    spike if spike_ref.dtype == jnp.bool_ else spike > 0.,
                    lambda out: out + weight_col,
                    lambda out: out,
                    temp
                )

            i_row_out = jax.lax.fori_loop(
                0,
                spike_ref.shape[0],
                loop_fn,
                jnp.zeros((mat_block_dim,), dtype=weight_ref.dtype)
            )
            out_ref[pl.ds(i_row_start, mat_block_dim)] = jnp.where(i_row_mask, i_row_out, 0.0)

        def run(weights, spikes):
            fn = pl.pallas_call(kernel, grid=(cdiv(m, mat_block_dim),), out_shape=kwargs['outs'], backend='triton')
            return fn(weights, spikes)

    return run


def _binary_densemv_jvp_weights(w_dot, weights, spikes, *, transpose, **kwargs):
    return binary_densemv_p_call(w_dot, spikes, transpose=transpose, backend=kwargs['backend'])


def _binary_densemv_jvp_spikes(spk_dot, weights, spikes, *, transpose, **kwargs):
    if transpose:
        return [spk_dot @ weights]
    else:
        return [weights @ spk_dot]


def _binary_densemv_transpose_rule(ct, weights, spikes, *, transpose, **kwargs):
    ct = ct[0]
    if ad.is_undefined_primal(spikes):
        if transpose:
            ct_spikes = jnp.matmul(weights, ct)
        else:
            ct_spikes = jnp.matmul(ct, weights)
        return weights, (ad.Zero(spikes) if type(ct) is ad.Zero else ct_spikes)
    else:
        if transpose:
            ct_weights = jnp.outer(spikes, ct)
        else:
            ct_weights = jnp.outer(ct, spikes)
        return (ad.Zero(weights) if type(ct) is ad.Zero else ct_weights), spikes


def _binary_densemv_batching(args, axes, *, transpose, **kwargs):
    if transpose:
        # weights[k,n], spikes[k] -> out[n]
        # mm transpose=True: weights[k,m].T @ spikes[k,n] -> out[m,n]
        # result shape from mm: [weights.shape[1], spikes.shape[1]]
        if axes == (None, 0):
            # spikes batched on axis 0: [batch, k] -> .T gives [k, batch]
            # mm(weights[k,n], [k,batch], transpose=True) -> [n, batch]
            r = binary_densemm(args[0], args[1].T, transpose=True, backend=kwargs['backend'])
            return [r], [1]
        if axes == (None, 1):
            # spikes batched on axis 1: [k, batch]
            # mm(weights[k,n], [k,batch], transpose=True) -> [n, batch]
            r = binary_densemm(args[0], args[1], transpose=True, backend=kwargs['backend'])
            return [r], [1]
    else:
        # weights[m,k], spikes[k] -> out[m]
        if axes == (None, 0):
            r = binary_densemm(args[0], args[1].T, transpose=False, backend=kwargs['backend'])
            return [r], [1]
        if axes == (None, 1):
            r = binary_densemm(args[0], args[1], transpose=False, backend=kwargs['backend'])
            return [r], [1]
    return general_batching_rule(binary_densemv_p, args, axes, transpose=transpose, **kwargs)


def _binary_densemv_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for bool_event in (True, False):
        # transpose=False benchmark (dbmv style)
        weights = jnp.asarray(np.random.randn(n_pre, n_post), dtype=dtype)
        if bool_event:
            spikes = jnp.asarray(np.random.rand(n_post) > (1 - prob), dtype=jnp.bool_)
        else:
            spikes = jnp.asarray(np.random.rand(n_post), dtype=dtype)
        name = f"{'bool' if bool_event else 'float'}"
        configs.append(BenchmarkConfig(name, (weights, spikes)))
    return configs


def binary_densemv_p_call(weights, spikes, *, transpose, backend: Optional[str] = None):
    """
    Low-level primitive call for binary dense matrix-vector multiplication.

    This function validates input shapes, constructs the output shape
    descriptor, and invokes the ``binary_densemv_p`` JAX primitive. Unlike
    :func:`binary_densemv`, this function operates on raw numerical arrays
    without ``brainunit`` unit handling.

    Parameters
    ----------
    weights : jax.Array
        The weight matrix. Shape ``(m, k)`` when ``transpose=False``,
        or ``(k, n)`` when ``transpose=True``.
    spikes : jax.Array
        The binary vector with shape ``(k,)``. Can be boolean or float.
    transpose : bool
        If False, compute ``weights @ spikes`` producing shape ``(m,)``.
        If True, compute ``spikes @ weights`` producing shape ``(n,)``.
    backend : str, optional
        Backend to use for the computation. One of ``'numba'``, ``'warp'``,
        ``'pallas'``, or ``None`` (auto-select).

    Returns
    -------
    result : list of jax.Array
        A single-element list containing the result vector. Shape ``(m,)``
        when ``transpose=False``, or ``(n,)`` when ``transpose=True``.

    Raises
    ------
    AssertionError
        If the inner dimensions of ``weights`` and ``spikes`` do not match.

    See Also
    --------
    binary_densemv : High-level function with unit handling.

    Notes
    -----
    This is the low-level entry point that bypasses unit handling. The
    mathematical operation is identical to :func:`binary_densemv`:

    When ``transpose=False``:

    ``out[i] = sum_{j where s[j] active} weights[i, j]``

    When ``transpose=True``:

    ``out[j] = sum_{i where s[i] active} weights[i, j]``

    The function returns a single-element list to conform to the JAX
    primitive output convention.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._dense.binary import binary_densemv_p_call
        >>> weights = jnp.ones((3, 4), dtype=jnp.float32)
        >>> spikes = jnp.array([1., 0., 1., 0.], dtype=jnp.float32)
        >>> binary_densemv_p_call(weights, spikes, transpose=False)
    """
    if transpose:
        assert spikes.shape[0] == weights.shape[0], (
            f"shapes {spikes.shape} and {weights.shape} not aligned: "
            f"{spikes.shape[0]} (dim 0) != {weights.shape[0]} (dim 0)"
        )
        out = jax.ShapeDtypeStruct([weights.shape[1]], weights.dtype)
    else:
        assert spikes.shape[0] == weights.shape[1], (
            f"spikes shape {spikes.shape} and weights shape {weights.shape} are not compatible"
        )
        out = jax.ShapeDtypeStruct([weights.shape[0]], weights.dtype)
    return binary_densemv_p(
        weights,
        spikes,
        outs=[out],
        transpose=transpose,
        spk_info=jax.ShapeDtypeStruct(spikes.shape, spikes.dtype),
        weight_info=jax.ShapeDtypeStruct(weights.shape, weights.dtype),
        backend=backend,
    )


binary_densemv_p = XLACustomKernel(
    'binary_densemv',
    doc="""
Low-level XLA custom-kernel primitive for ``binary_densemv``.

This ``XLACustomKernel`` instance dispatches the binary (event-driven) dense
matrix-vector multiplication operation to registered backends (``numba``, ``warp``,
``pallas``), using runtime shape/dtype metadata provided by the high-level wrapper.

The operation computes ``weights[m,k] @ spikes[k] -> out[m]`` when ``transpose=False``,
or ``spikes[k] @ weights[k,n] -> out[n]`` when ``transpose=True``, where only active
(nonzero) spike entries contribute to the output.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``binary_densemv_p.available_backends(platform)``,
and the default backend can be configured with ``binary_densemv_p.set_default(platform, backend)``.

See Also
--------
binary_densemv : High-level user-facing function wrapper.
"""
)
binary_densemv_p.def_numba_kernel(_binary_densemv_numba_kernel)
binary_densemv_p.def_warp_kernel(_binary_densemv_warp_kernel)
binary_densemv_p.def_pallas_kernel('gpu', _binary_densemv_pallas_kernel)
binary_densemv_p.def_jvp_rule2(_binary_densemv_jvp_weights, _binary_densemv_jvp_spikes)
binary_densemv_p.def_transpose_rule(_binary_densemv_transpose_rule)
binary_densemv_p.def_batching_rule(_binary_densemv_batching)
binary_densemv_p.def_call(binary_densemv_p_call)
binary_densemv_p.def_tags('dense', 'binary')
binary_densemv_p.def_benchmark_data(_binary_densemv_benchmark_data)


# ==============================================================================
# Unified binary dense matrix-matrix product (binary_densemm)
# ==============================================================================
#
# transpose=False: weights[m,k] @ spikes[k,n] -> out[m,n]  (old dbmm)
# transpose=True:  weights[k,m].T @ spikes[k,n] -> out[m,n]  (old bdmm)
#
# Argument order is always (weights, spikes).


@namescope(static_argnames=['transpose'])
def binary_densemm(weights, spikes, *, transpose, backend: Optional[str] = None):
    """
    Perform event-driven dense matrix-matrix multiplication with binary spikes.

    When ``transpose=False``, computes ``weights[m, k] @ spikes[k, n] -> out[m, n]``
    (dense matrix times binary matrix).

    When ``transpose=True``, computes ``weights[k, m].T @ spikes[k, n] -> out[m, n]``
    (transposed dense matrix times binary matrix). Both weights and spikes share
    their first dimension ``k``.

    Parameters
    ----------
    weights : array_like
        The weight matrix. Shape ``(m, k)`` when ``transpose=False``,
        or ``(k, m)`` when ``transpose=True``. Can be a ``brainunit`` quantity.
    spikes : array_like
        The binary matrix. Shape ``(k, n)`` in both modes. Can be boolean or
        float. Can be a ``brainunit`` quantity.
    transpose : bool
        If False, compute ``weights @ spikes``. If True, compute
        ``weights.T @ spikes``.
    backend : str, optional
        Backend to use for the computation. One of ``'numba'``, ``'warp'``,
        ``'pallas'``, or ``None`` (auto-select).

    Returns
    -------
    result : array_like
        Result matrix with shape ``(m, n)``. If inputs carry units, the
        result carries the product of the weight and spike units.

    Raises
    ------
    AssertionError
        If the shared dimensions of ``weights`` and ``spikes`` do not match.

    See Also
    --------
    binary_densemv : Matrix-vector variant of binary dense multiplication.

    Notes
    -----
    When ``transpose=False``, the operation computes:

    ``out[i, j] = sum_{k} W[i, k] * s[k, j]``

    where ``s[k, j]`` is ``1`` if ``spikes[k, j]`` is active and ``0``
    otherwise.

    When ``transpose=True``, the operation computes:

    ``out[i, j] = sum_{k} W[k, i] * s[k, j]``

    where ``s[k, j]`` is ``1`` if ``spikes[k, j]`` is active and ``0``
    otherwise.

    Boolean spikes are converted to ``weights.dtype`` before the primitive
    call to avoid Pallas Triton boolean buffer corruption.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._dense.binary import binary_densemm
        >>> weights = jnp.ones((3, 4), dtype=jnp.float32)
        >>> spikes = jnp.array([[True, False],
        ...                     [False, True],
        ...                     [True, True],
        ...                     [False, False]])
        >>> binary_densemm(weights, spikes, transpose=False)
        Array([[2., 2.],
               [2., 2.],
               [2., 2.]], dtype=float32)
    """
    with jax.ensure_compile_time_eval():
        weights = u.math.asarray(weights)
        spikes = u.math.asarray(spikes)
    weight_val, wunit = u.split_mantissa_unit(weights)
    spk_val, spkunit = u.split_mantissa_unit(spikes)
    r = binary_densemm_p_call(weight_val, spk_val, transpose=transpose, backend=backend)
    return u.maybe_decimal(r[0] * wunit * spkunit)


def _binary_densemm_numba_kernel(
    spk_info: jax.ShapeDtypeStruct,
    weight_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import numba

    if transpose:
        # weights[k,m].T @ spikes[k,n] -> out[m,n]
        # primitive args: (weights, spikes)
        if spk_info.dtype == jnp.bool_:
            @numba.njit(parallel=get_numba_parallel(), fastmath=True, nogil=True)
            def kernel(weights, spikes, posts):
                for i_n in numba.prange(spikes.shape[1]):
                    out = np.zeros(weights.shape[1], dtype=posts.dtype)
                    for i_k in range(spikes.shape[0]):
                        if spikes[i_k, i_n]:
                            out += weights[i_k]
                    posts[:, i_n] = out
        else:
            @numba.njit(parallel=get_numba_parallel(), fastmath=True, nogil=True)
            def kernel(weights, spikes, posts):
                for i_n in numba.prange(spikes.shape[1]):
                    out = np.zeros(weights.shape[1], dtype=posts.dtype)
                    for i_k in range(spikes.shape[0]):
                        if spikes[i_k, i_n] > 0.:
                            out += weights[i_k]
                    posts[:, i_n] = out
    else:
        # weights[m,k] @ spikes[k,n] -> out[m,n]
        if spk_info.dtype == jnp.bool_:
            @numba.njit(parallel=get_numba_parallel(), fastmath=True, nogil=True)
            def kernel(weights, spikes, posts):
                for i_n in numba.prange(spikes.shape[1]):
                    out = np.zeros(weights.shape[0], dtype=weights.dtype)
                    for i_k in range(spikes.shape[0]):
                        if spikes[i_k, i_n]:
                            out += weights[:, i_k]
                    posts[:, i_n] = out
        else:
            @numba.njit(parallel=get_numba_parallel(), fastmath=True, nogil=True)
            def kernel(weights, spikes, posts):
                for i_n in numba.prange(spikes.shape[1]):
                    out = np.zeros(weights.shape[0], dtype=weights.dtype)
                    for i_k in range(spikes.shape[0]):
                        if spikes[i_k, i_n] > 0.:
                            out += weights[:, i_k]
                    posts[:, i_n] = out

    def run(weights, spikes):
        return numba_kernel(kernel, outs=kwargs['outs'])(weights, spikes)

    return run


def _binary_densemm_warp_kernel(
    weight_info: jax.ShapeDtypeStruct,
    spk_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    weight_warp_info = jaxinfo_to_warpinfo(weight_info)
    out_warp_info = jaxinfo_to_warpinfo(kwargs['outs'][0])

    if transpose:
        # weights[k,m].T @ spikes[k,n] -> out[m,n]
        k = weight_info.shape[0]
        m = weight_info.shape[1]
        n = spk_info.shape[1]

        if spk_info.dtype == jnp.bool_:
            spk_float_info = jax.ShapeDtypeStruct(spk_info.shape, jnp.float32)
            spike_warp_info = jaxinfo_to_warpinfo(spk_float_info)

            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info,
            ):
                i_m, i_n = warp.tid()
                r = weights.dtype(0.)
                for i_k in range(k):
                    if spikes[i_k, i_n] > 0.:
                        r += weights[i_k, i_m]
                out[i_m, i_n] = r

            def run(weights, spikes):
                spikes = spikes.astype(jnp.float32)
                out_info = kwargs['outs'][0]
                fn = jax_kernel(kernel, launch_dims=(m, n), num_outputs=1, output_dims={'out': out_info.shape})
                return fn(weights, spikes)
        else:
            spike_warp_info = jaxinfo_to_warpinfo(spk_info)

            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info,
            ):
                i_m, i_n = warp.tid()
                r = weights.dtype(0.)
                for i_k in range(k):
                    if spikes[i_k, i_n] > 0.:
                        r += weights[i_k, i_m]
                out[i_m, i_n] = r

            def run(weights, spikes):
                out_info = kwargs['outs'][0]
                fn = jax_kernel(kernel, launch_dims=(m, n), num_outputs=1, output_dims={'out': out_info.shape})
                return fn(weights, spikes)

    else:
        # weights[m,k] @ spikes[k,n] -> out[m,n]
        k = spk_info.shape[0]
        n = spk_info.shape[1]
        m = weight_info.shape[0]

        if spk_info.dtype == jnp.bool_:
            spk_float_info = jax.ShapeDtypeStruct(spk_info.shape, jnp.float32)
            spike_warp_info = jaxinfo_to_warpinfo(spk_float_info)

            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info
            ):
                i_m, i_n = warp.tid()
                r = weights.dtype(0.)
                for i_k in range(k):
                    if spikes[i_k, i_n] > 0.:
                        r += weights[i_m, i_k]
                out[i_m, i_n] = r

            def run(weights, spikes):
                spikes = spikes.astype(jnp.float32)
                out_info = kwargs['outs'][0]
                fn = jax_kernel(kernel, launch_dims=(m, n), num_outputs=1, output_dims={'out': out_info.shape})
                return fn(weights, spikes)
        else:
            spike_warp_info = jaxinfo_to_warpinfo(spk_info)

            @warp.kernel
            def kernel(
                weights: weight_warp_info,
                spikes: spike_warp_info,
                out: out_warp_info
            ):
                i_m, i_n = warp.tid()
                r = weights.dtype(0.)
                for i_k in range(k):
                    if spikes[i_k, i_n] > 0.:
                        r += weights[i_m, i_k]
                out[i_m, i_n] = r

            def run(weights, spikes):
                out_info = kwargs['outs'][0]
                fn = jax_kernel(kernel, launch_dims=(m, n), num_outputs=1, output_dims={'out': out_info.shape})
                return fn(weights, spikes)

    return run


def _binary_densemm_pallas_kernel(
    weight_info: jax.ShapeDtypeStruct,
    spk_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    from jax.experimental import pallas as pl

    if transpose:
        # weights[k,m].T @ spikes[k,n] -> out[m,n]
        k = weight_info.shape[0]
        m = weight_info.shape[1]
        n = spk_info.shape[1]

        def kernel(
            weight_ref,  # [k, m]
            spike_ref,  # [k, n]
            out_ref,  # [m, n]
        ):
            i_n = pl.program_id(0)
            i_m = pl.program_id(1)

            def loop_fn(i_k, temp):
                spike = spike_ref[i_k, i_n]
                weight = weight_ref[i_k, i_m]
                return temp + jnp.where(spike > 0., weight, 0.0)

            out_ref[i_m, i_n] = jax.lax.fori_loop(
                0, k, loop_fn, jnp.zeros((), dtype=weight_ref.dtype)
            )

        def run(weights, spikes):
            fn = pl.pallas_call(kernel, grid=(n, m), out_shape=kwargs['outs'], backend='triton')
            return fn(weights, spikes)

    else:
        # weights[m,k] @ spikes[k,n] -> out[m,n]
        k = spk_info.shape[0]
        n = spk_info.shape[1]
        m = weight_info.shape[0]

        def kernel(
            weight_ref,  # [m, k]
            spike_ref,  # [k, n]
            out_ref,  # [m, n]
        ):
            i_n = pl.program_id(0)
            i_m = pl.program_id(1)

            def loop_fn(i_k, temp):
                spike = spike_ref[i_k, i_n]
                weight = weight_ref[i_m, i_k]
                return temp + jnp.where(spike > 0., weight, 0.0)

            out_ref[i_m, i_n] = jax.lax.fori_loop(
                0, k, loop_fn, jnp.zeros((), dtype=weight_ref.dtype)
            )

        def run(weights, spikes):
            fn = pl.pallas_call(kernel, grid=(n, m), out_shape=kwargs['outs'], backend='triton')
            return fn(weights, spikes)

    return run


def _binary_densemm_jvp_weights(w_dot, weights, spikes, *, transpose, **kwargs):
    return binary_densemm_p_call(w_dot, spikes, transpose=transpose, backend=kwargs['backend'])


def _binary_densemm_jvp_spikes(spk_dot, weights, spikes, *, transpose, **kwargs):
    if transpose:
        return [weights.T @ spk_dot]
    else:
        return [weights @ spk_dot]


def _binary_densemm_transpose_rule(ct, weights, spikes, *, transpose, **kwargs):
    ct = ct[0]
    if transpose:
        # weights[k,m].T @ spikes[k,n] -> out[m,n]
        if ad.is_undefined_primal(spikes):
            # ct_spikes: weights[k,m] @ ct[m,n] -> [k,n]
            ct_events = weights @ ct
            return weights, (ad.Zero(spikes) if type(ct) is ad.Zero else ct_events)
        else:
            # ct_weights: spikes[k,n] @ ct.T[n,m] -> [k,m]
            ct_weights = spikes @ ct.T
            return (ad.Zero(weights) if type(ct) is ad.Zero else ct_weights), spikes
    else:
        # weights[m,k] @ spikes[k,n] -> out[m,n]
        if ad.is_undefined_primal(spikes):
            ct_events = weights.T @ ct
            return weights, (ad.Zero(spikes) if type(ct) is ad.Zero else ct_events)
        else:
            # ct[m,n], spikes[k,n] -> ct_weights[m,k]
            # ct[m,n] @ spikes.T[n,k] -> [m,k]
            ct_weights = binary_densemm(ct, spikes.T, transpose=False, backend=kwargs['backend'])
            return (ad.Zero(weights) if type(ct) is ad.Zero else ct_weights), spikes


def _binary_densemm_batching_spikes_fn(args, axes, *, transpose, **kwargs):
    weights, spikes = args
    if transpose:
        # weights[k,m].T @ spikes[k,n] -> out[m,n]
        # out shape: [weights.shape[1], spikes.shape[1]]
        # spikes batched: [batch,k,n] or [k,batch,n] or [k,n,batch]
        assert spikes.ndim == 3, 'requires 3D events.'
        assert weights.ndim == 2, 'requires 2D weights.'
        spk_axis = axes[1]
        if spk_axis == 0:
            # [batch,k,n] -> transpose(1,0,2) -> [k,batch,n] -> reshape [k, batch*n]
            spikes = jnp.transpose(spikes, (1, 0, 2))
            k, batch, n_val = spikes.shape
            events = spikes.reshape(k, batch * n_val)
            r = binary_densemm_p_call(weights, events, transpose=True, backend=kwargs['backend'])
            # result: [m, batch*n] -> reshape [m, batch, n]
            r = jnp.reshape(r[0], [r[0].shape[0], batch, n_val])
            return [r], [1]
        elif spk_axis == 1:
            # [k,batch,n] -> reshape [k, batch*n]
            k, batch, n_val = spikes.shape
            events = spikes.reshape(k, batch * n_val)
            r = binary_densemm_p_call(weights, events, transpose=True, backend=kwargs['backend'])
            # result: [m, batch*n] -> reshape [m, batch, n]
            r = jnp.reshape(r[0], [r[0].shape[0], batch, n_val])
            return [r], [1]
        elif spk_axis == 2:
            # [k,n,batch] -> reshape [k, n*batch]
            k, n_val, batch = spikes.shape
            events = spikes.reshape(k, n_val * batch)
            r = binary_densemm_p_call(weights, events, transpose=True, backend=kwargs['backend'])
            # result: [m, n*batch] -> reshape [m, n, batch]
            r = jnp.reshape(r[0], [r[0].shape[0], n_val, batch])
            return [r], [2]
    else:
        # weights[m,k] @ spikes[k,n] -> out[m,n]
        # spikes batched
        assert spikes.ndim == 3, 'requires 3D events.'
        assert weights.ndim == 2, 'requires 2D weights.'
        spk_axis = axes[1]
        if spk_axis == 0:
            spikes = jnp.transpose(spikes, (1, 0, 2))
            k, maybe_batch1, maybe_batch2 = spikes.shape
            events = spikes.reshape(k, maybe_batch1 * maybe_batch2)
            r = binary_densemm_p_call(weights, events, transpose=False, backend=kwargs['backend'])
            r = jnp.reshape(r[0], [r[0].shape[0], maybe_batch1, maybe_batch2])
            return [r], [1]
        elif spk_axis == 1:
            k, maybe_batch1, maybe_batch2 = spikes.shape
            events = spikes.reshape(k, maybe_batch1 * maybe_batch2)
            r = binary_densemm_p_call(weights, events, transpose=False, backend=kwargs['backend'])
            r = jnp.reshape(r[0], [r[0].shape[0], maybe_batch1, maybe_batch2])
            return [r], [1]
        elif spk_axis == 2:
            k, maybe_batch1, maybe_batch2 = spikes.shape
            events = spikes.reshape(k, maybe_batch1 * maybe_batch2)
            r = binary_densemm_p_call(weights, events, transpose=False, backend=kwargs['backend'])
            r = jnp.reshape(r[0], [r[0].shape[0], maybe_batch1, maybe_batch2])
            return [r], [2]
    return general_batching_rule(binary_densemm_p, args, axes, transpose=transpose, **kwargs)


def _binary_densemm_batching_weights_fn(args, axes, *, transpose, **kwargs):
    weights, spikes = args
    if transpose:
        # weights[k,m].T @ spikes[k,n] -> out[m,n]
        # out shape: [weights.shape[1], spikes.shape[1]]
        # weights batched: [batch,k,m] or [k,batch,m] or [k,m,batch]
        assert weights.ndim == 3, 'requires 3D weights.'
        assert spikes.ndim == 2, 'requires 2D events.'
        w_axis = axes[0]
        if w_axis == 0:
            # [batch,k,m] -> transpose(1,0,2) -> [k,batch,m] -> reshape [k, batch*m]
            weights = jnp.transpose(weights, (1, 0, 2))
            k, batch, m_val = weights.shape
            w = weights.reshape(k, batch * m_val)
            r = binary_densemm_p_call(w, spikes, transpose=True, backend=kwargs['backend'])
            # result: [batch*m, n] -> reshape [batch, m, n]
            r = jnp.reshape(r[0], [batch, m_val, r[0].shape[1]])
            return [r], [0]
        elif w_axis == 1:
            # [k,batch,m] -> reshape [k, batch*m]
            k, batch, m_val = weights.shape
            w = weights.reshape(k, batch * m_val)
            r = binary_densemm_p_call(w, spikes, transpose=True, backend=kwargs['backend'])
            # result: [batch*m, n] -> reshape [batch, m, n]
            r = jnp.reshape(r[0], [batch, m_val, r[0].shape[1]])
            return [r], [0]
        elif w_axis == 2:
            # [k,m,batch] -> reshape [k, m*batch]
            k, m_val, batch = weights.shape
            w = weights.reshape(k, m_val * batch)
            r = binary_densemm_p_call(w, spikes, transpose=True, backend=kwargs['backend'])
            # result: [m*batch, n] -> reshape [m, batch, n]
            r = jnp.reshape(r[0], [m_val, batch, r[0].shape[1]])
            return [r], [1]
    else:
        # weights[m,k] @ spikes[k,n] -> out[m,n]
        # weights batched
        assert weights.ndim == 3, 'requires 3D weights.'
        assert spikes.ndim == 2, 'requires 2D events.'
        w_axis = axes[0]
        if w_axis == 0:
            maybe_batch1, maybe_batch2, k = weights.shape
            w = weights.reshape(maybe_batch1 * maybe_batch2, k)
            r = binary_densemm_p_call(w, spikes, transpose=False, backend=kwargs['backend'])
            r = jnp.reshape(r[0], [maybe_batch1, maybe_batch2, r[0].shape[-1]])
            return [r], [0]
        elif w_axis == 1:
            maybe_batch1, maybe_batch2, k = weights.shape
            w = weights.reshape(maybe_batch1 * maybe_batch2, k)
            r = binary_densemm_p_call(w, spikes, transpose=False, backend=kwargs['backend'])
            r = jnp.reshape(r[0], [maybe_batch1, maybe_batch2, r[0].shape[-1]])
            return [r], [1]
        elif w_axis == 2:
            weights = jnp.transpose(weights, (0, 2, 1))
            maybe_batch1, maybe_batch2, k = weights.shape
            w = weights.reshape(maybe_batch1 * maybe_batch2, k)
            r = binary_densemm_p_call(w, spikes, transpose=False, backend=kwargs['backend'])
            r = jnp.reshape(r[0], [maybe_batch1, maybe_batch2, r[0].shape[-1]])
            return [r], [1]
    return general_batching_rule(binary_densemm_p, args, axes, transpose=transpose, **kwargs)


def _binary_densemm_batching(args, axes, *, transpose, **kwargs):
    w_axis, spk_axis = axes
    if w_axis is None:
        return _binary_densemm_batching_spikes_fn(args, axes, transpose=transpose, **kwargs)
    elif spk_axis is None:
        return _binary_densemm_batching_weights_fn(args, axes, transpose=transpose, **kwargs)
    else:
        return general_batching_rule(binary_densemm_p, args, axes, transpose=transpose, **kwargs)


def _binary_densemm_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for bool_event in (True, False):
        weights = jnp.asarray(np.random.randn(n_pre, n_post), dtype=dtype)
        if bool_event:
            spikes = jnp.asarray(np.random.rand(n_post, 10) > (1 - prob), dtype=jnp.bool_)
        else:
            spikes = jnp.asarray(np.random.rand(n_post, 10), dtype=dtype)
        name = f"{'bool' if bool_event else 'float'}"
        configs.append(BenchmarkConfig(name, (weights, spikes)))
    return configs


def binary_densemm_p_call(weights, spikes, *, transpose, backend: Optional[str] = None):
    """
    Low-level primitive call for binary dense matrix-matrix multiplication.

    This function validates input shapes, converts boolean spikes to float
    to avoid Pallas Triton boolean buffer corruption, constructs the output
    shape descriptor, and invokes the ``binary_densemm_p`` JAX primitive.
    Unlike :func:`binary_densemm`, this function operates on raw numerical
    arrays without ``brainunit`` unit handling.

    Parameters
    ----------
    weights : jax.Array
        The weight matrix. Shape ``(m, k)`` when ``transpose=False``,
        or ``(k, m)`` when ``transpose=True``.
    spikes : jax.Array
        The binary matrix with shape ``(k, n)``. Can be boolean or float.
        Boolean inputs are automatically cast to ``weights.dtype``.
    transpose : bool
        If False, compute ``weights @ spikes`` producing shape ``(m, n)``.
        If True, compute ``weights.T @ spikes`` producing shape ``(m, n)``.
    backend : str, optional
        Backend to use for the computation. One of ``'numba'``, ``'warp'``,
        ``'pallas'``, or ``None`` (auto-select).

    Returns
    -------
    result : list of jax.Array
        A single-element list containing the result matrix with shape
        ``(m, n)``.

    Raises
    ------
    AssertionError
        If the shared dimensions of ``weights`` and ``spikes`` do not match.

    See Also
    --------
    binary_densemm : High-level function with unit handling.

    Notes
    -----
    This is the low-level entry point that bypasses unit handling. The
    mathematical operation is identical to :func:`binary_densemm`:

    When ``transpose=False``:

    ``out[i, j] = sum_{k where s[k, j] active} weights[i, k]``

    When ``transpose=True``:

    ``out[i, j] = sum_{k where s[k, j] active} weights[k, i]``

    Boolean spikes are automatically cast to ``weights.dtype`` before
    invoking the primitive to prevent Pallas Triton boolean buffer
    corruption. The function returns a single-element list to conform to
    the JAX primitive output convention.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._dense.binary import binary_densemm_p_call
        >>> weights = jnp.ones((3, 4), dtype=jnp.float32)
        >>> spikes = jnp.array([[1., 0.], [0., 1.],
        ...                     [1., 1.], [0., 0.]], dtype=jnp.float32)
        >>> binary_densemm_p_call(weights, spikes, transpose=False)
    """
    # Convert bool spikes to float before primitive call to avoid
    # Pallas Triton boolean buffer corruption
    if spikes.dtype == jnp.bool_:
        spikes = jnp.asarray(spikes, dtype=weights.dtype)
    if transpose:
        # weights[k,m].T @ spikes[k,n] -> out[m,n]
        assert weights.shape[0] == spikes.shape[0], (
            f"weights shape {weights.shape} and spikes shape {spikes.shape} do not match"
            f" for event matrix multiplication: weights dim 0 ({weights.shape[0]}) != spikes dim 0 ({spikes.shape[0]})"
        )
        out = jax.ShapeDtypeStruct([weights.shape[1], spikes.shape[1]], weights.dtype)
    else:
        # weights[m,k] @ spikes[k,n] -> out[m,n]
        assert weights.shape[1] == spikes.shape[0], (
            f"weights.shape[1] ({weights.shape[1]}) != spikes.shape[0] ({spikes.shape[0]})"
            f", weights: {weights.shape}, spikes: {spikes.shape}"
        )
        out = jax.ShapeDtypeStruct([weights.shape[0], spikes.shape[1]], weights.dtype)
    return binary_densemm_p(
        weights,
        spikes,
        outs=[out],
        transpose=transpose,
        spk_info=jax.ShapeDtypeStruct(spikes.shape, spikes.dtype),
        weight_info=jax.ShapeDtypeStruct(weights.shape, weights.dtype),
        backend=backend,
    )


binary_densemm_p = XLACustomKernel(
    'binary_densemm',
    doc="""
Low-level XLA custom-kernel primitive for ``binary_densemm``.

This ``XLACustomKernel`` instance dispatches the binary (event-driven) dense
matrix-matrix multiplication operation to registered backends (``numba``, ``warp``,
``pallas``), using runtime shape/dtype metadata provided by the high-level wrapper.

The operation computes ``weights[m,k] @ spikes[k,n] -> out[m,n]`` when ``transpose=False``,
or ``weights[k,m].T @ spikes[k,n] -> out[m,n]`` when ``transpose=True``, where only active
(nonzero) spike entries contribute to the output.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``binary_densemm_p.available_backends(platform)``,
and the default backend can be configured with ``binary_densemm_p.set_default(platform, backend)``.

See Also
--------
binary_densemm : High-level user-facing function wrapper.
"""
)
binary_densemm_p.def_numba_kernel(_binary_densemm_numba_kernel)
binary_densemm_p.def_warp_kernel(_binary_densemm_warp_kernel)
binary_densemm_p.def_pallas_kernel('gpu', _binary_densemm_pallas_kernel)
binary_densemm_p.def_jvp_rule2(_binary_densemm_jvp_weights, _binary_densemm_jvp_spikes)
binary_densemm_p.def_transpose_rule(_binary_densemm_transpose_rule)
binary_densemm_p.def_batching_rule(_binary_densemm_batching)
binary_densemm_p.def_call(binary_densemm_p_call)
binary_densemm_p.def_tags('dense', 'binary')
binary_densemm_p.def_benchmark_data(_binary_densemm_benchmark_data)
